# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""A collection of doc strings shared between grammar rules."""




EXPRESSION_ONLY_DOC = """Parses python expressions outside of a pipeline.

Examples:
    1. x = 5
    2. x = y % (5.7 + z / 2)
    3. x = w and (y or not z)

Grammar:
    !expr

Args:
    expr (expression): The expression to be evaluated.

Type:
    Function
"""


ON_DF_DOC = """Starts a pipeline on the given DataFrame.

Examples:
    1. x = on df
    2. x = on df | transformer 1 ... | transformer n ... | operation

Grammar:
    on $dataframe

Args:
    dataframe (variable): The name of DataFrame

Type:
    Initialization
"""


CREATE_DF_DOC = """Creates a new DataFrame from an list.

Examples:
    1. x = create dataframe from my_data with header 'col1', 'col2', 'col3'

Grammar:
    create dataframe from $data with header $header[$col_name]

Args:
    data (variable): The data from which to create the dataframe.
    header (list): A list of column names

Type:
    Initialization
"""


LOAD_FROM_DOC = """Load a DataFrame from a file.

Examples:
    1. x = load from "my_file.json" as json
    2. x = load from "my_file.csv" as csv | drop duplicates

Grammar:
    load from $path as $type
    type := { json, csv }

Args:
    path (variable): A string containing the path to a file.
    type (variable): The type of the file.

Type:
    Initialization
"""


SAVE_TO_DOC = """Save a DataFrame to a file.

Examples:
    1. on df | save to "my_file.json" as json
    2. on df | save to "my_file.csv" as csv

Grammar:
    save to $path as $type
    type := { json, csv }

Args:
    path (variable): A string containing the path to a file.
    type (variable): The type of the file.

Type:
    Operation
"""


UNION_DOC = """Compute the union of rows.

Examples:
    1. x = on df | union other

Grammar:
    union $table

Args:
    table (variable): The table from which all rows will be added.

Type:
    Transformation
"""


DIFFERENCE_DOC = """Remove all rows which are contained in `table`.

Examples:
    1. x = on df | difference other

Grammar:
    difference $table

Args:
    table (variable): The table which contains all rows that shall be removed.

Type:
    Transformation
"""


INTERSECTION_DOC = """Remove all rows which are not contained in `table`.

Examples:
    1. x = on df | intersection other

Grammar:
    intersection $table

Args:
    table (variable): The table which contains all rows that shall not be removed.

Type:
    Transformation
"""


JOIN_DOC = """Compute join with this DataFrame.

Examples:
    1. x = on df | join inner df2 on 'col1', 'col2'
    2. x = on df | join left df2 on 'col1'

Grammar:
    join $how $table on $columns[$col]
    how := { left, right, outer, inner }

Args:
    how (variable): How the join shall be performed.
    table (variable): The table with which to join.
    columns (varlist): A list of column on which to join.

Type:
    Transformation
"""


SELECT_COLUMNS_DOC = """Select certain columns from a DataFrame.

Examples:
    1. ## x = on df | select columns df.col1, col2, df.col3
    2. ## x = on df | select columns "col1", "col2", "col3"

Grammar:
    select columns $columns[$col]

Args:
    columns (varlist): A list of column names.

Type:
    Transformation
"""


SELECT_ROWS_DOC = """Select the rows of a DataFrame based on some condition.

The condition can be composed out of boolean, comparison and arithmetic
expression. The operator precedence is equivalent to python and it is possible
to use brackets to modify it.

Examples:
    1. ## x = on df | select rows df.col1 > (14.2 + z) and df.col2 == 'A'
    2. ## x = on df | select rows df.col1 != 0 and not df.col2 in [3, 5, 7]
    3. ## x = on df | select rows df.col3 % df.col1 != 2 or df.col1 <= 12

Grammar:
    select rows !condition

Args:
    condition (expression): A boolean expression used as a row filter.

Type:
    Transformation
"""


DROP_COLUMNS_DOC = """Drop certain columns from a DataFrame.

Examples:
    1. ## x = on df | drop columns df.col1, col2, df.col3
    2. ## x = on df | drop columns "col1", "col2", "col3"

Grammar:
    drop columns $columns[$col]

Args:
    columns (varlist): A list of column names.

Type:
    Transformation

"""


GROUP_BY_DOC = """Group a DataFrame and apply an aggregation.

Examples:
    1. ## x = on df | group by df.col1 apply min
    2. ## x = on df | group by df.col1, df.col2 apply mean

Grammar:
    group by $columns[$col] apply $aggregation
    aggregation := { min, max, sum, avg, mean, count }

Args:
    columns (varlist): A list of column names.
    aggregation (variable): The aggregation operation to be performed.
Type:
    Operation
"""


REPLACE_VALUES_DOC = """Replace a value with another.

Every occurrence of `old_value` will be substituted with `new_value`.

Examples:
    1. ## x = on df | replace values 1 by 0
    2. ## x = on df | replace values "old" by "new"

Grammar:
    replace $old_value with $new_value

Args:
    old_value (variable): The value to be replaced.
    new_value (variable): The value it will be replaced with.

Type:
    Operation
"""


APPEND_COLUMN_DOC = """Append a new column to the DataFrame.

Example:
    1. x = on df | append column df.col1 * 2 as 'new_col'

Grammar:
    append column !col_expr as $col_name

Args:
    col_expr (expression): An expression defining the new value of the column.
    col_name (variable): The new name of the column.

Type:
    Transformation
"""


SORT_BY_DOC = """Sort the DataFrame by certain columns.

Examples:
    1. ## x = on df | sort by df.col1 descending
    2. ## x = on df | sort by "col1" ascending, "col2" descending

Grammar:
    sort by $columns[$col $order]
    order := { ascending, descending }

Args:
    columns (varlist): A list of column names and sorting order pairs.

Type:
    Transformation
"""


DROP_DUPLICATES_DOC = """Drop duplicate rows from a DataFrame.

Examples:
    1. x = on df | drop duplicates
    2. x = on df | select rows df.col1 != 0 | drop duplicates

Grammar:
    drop duplicates

Type:
    Transformation
"""


RENAME_COLUMNS_DOC = """Rename some columns in a DataFrame.

Examples:
    1. x = on df | rename columns col1 to col2
    2. x = on df | rename columns col1 to col2, col3 to col4 | show

Grammar:
    rename columns $columns[$current to $new]

Args:
    columns (list): A list of current and new column names.

Type:
    Transformation
"""


SHOW_DOC = """Print the DataFrame to stdout.

Examples:
    1. on df | show
    2. on df | drop duplicates | show

Grammar:
    show

Type:
    Operation
"""


SHOW_SCHEMA_DOC = """Print the schema of the DataFrame to stdout.

Examples:
    1. on df | show schema

Grammar:
    show schema

Type:
    Operation
"""


DESCRIBE_DOC = """Print a description of the DataFrame to stdout.

Examples:
    1. on df | describe
    2. on df | drop duplicates | describe

Grammar:
    describe

Type:
    Operation
"""


HEAD_DOC = """Get the `num_rows` top most rows in the DataFrame.

Examples:
    1. on df | head 10
    2. on df | drop duplicates | head 100

Grammar:
    head $num_rows

Args:
    num_rows (variable): The number of rows to return.

Type:
    Operation
"""


COUNT_DOC = """Count the number of rows in the DataFrame.

Examples:
    1. on df | count
    2. on df | drop duplicates | count

Grammar:
    count

Type:
    Operation
"""



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""External Dynamic Grammar Rules are added via python functions.

It is advised to use the @grammar decorator for this, however it is not mandatory.
The following prototypes are valid:
* my_function(str: code, list: args, dict: env) -> str
* my_function(str: code, list: args) -> str
* my_function(str: code, dict: env) -> str
* my_function(str: code) -> str

In case my_function only takes two arguments the second argument must be named `args`or `env`.
Every class in this module contributes to the @grammar decorator.

Example:
    Implementation of a grammar rule for the pandas code generator::

        @grammar(PandasExpressionRule)
        def select_rows(code, args):
            \"\"\"Select the rows of a DataFrame based on some condition.

            The condition can be composed out of boolean, comparison and arithmetic
            expression. The operator precedence is equivalent to python and it is possible
            to use brackets to modify it.

            Examples:
                1. ## x = on df | select rows df.col1 > (14.2 + z) and df.col2 == 'A'
                2. ## x = on df | select rows df.col1 != 0 and not df.col2 in [3, 5, 7]
                3. ## x = on df | select rows df.col3 % df.col1 != 2 or df.col1 <= 12

            Grammar:
                select rows !condition

            Args:
                condition (expression): A boolean expression used as a row filter.

            Type:
                Transformer
            \"\"\"
            return code + "[{}]".format(args["condition"])
"""

from textwrap import dedent
from .types import Keyword, Variable, Expression, VarList
from .utils import decorator_factory, convert_function
from .rules import ExpressionRule, GrammarRule
from .exceptions import DSLArgumentError, DSLFunctionError




class GrammarParser:
    """A GrammarParser parses the argument list of a DSL sentence.
    The parsing procedure is defined via specification of the associated grammar.
    A grammar rule should be specified as follows::

        <name> <variable List | <variable> | <keyword> | <expression>...
        variable name 1> := { value_1, ... , value_n}
        ...
        <variable name n> := { value_1, ... , value_m}

    1. <name> can be any sequence of words (e.g. rename columns),
       optionally surrounded with brackets "(<name>)"
    2. <keyword> can be any word
    3. <variable> a word prefixed by the `$` sign (e.g $my_variable)
    4. <variable list> has the form $var_list_name[$variable1 keyword1 ...],
       any number of variables and keywords may be contained within the brackets.
       Providing a <var_list_name> is optional and will default to the tuple
       ($variable1, ... , $variable2)
    5. The rows of the form `<variable name> := { value_1, ... , value_n}` are optional
       and specify a finite set of values { value_1, ... , value_n} available to the variable.

    Example:
        Definition of the grammar rule "group by"::

            group by $columns[$col] apply $agg
            agg := { min, max, avg, sum }

    Args:
        input_str (str): A string containing grammar rules.
        expr_rule (ExpressionRule): The type of expression rule to be used
    """

    def __init__(self, input_str, expr_rule=ExpressionRule):
        """Please see help(GrammarParser)."""
        grammar_lines, choice_lines, function_type = self._parse_grammar_desc(input_str)
        grammar_desc = self._prepare_grammar_desc(grammar_lines)
        name, grammar_desc = self._extract_rule_name(grammar_desc)
        instructions = self._extract_instructions(grammar_desc)
        choices = self._extract_choices(choice_lines)
        self.name = name
        self.function_type = function_type
        self.rule = GrammarRule(instructions, choices, expr_rule)


    @property
    def grammar_rules(self):
        """The GrammarRule, responsible for argument conversion."""
        return self.rule.rules

    @property
    def grammar_choices(self):
        """Mapping from variable names to choices."""
        return self.rule.choices

    @property
    def grammar_rule_name(self):
        """The GrammarRule's name."""
        return self.name

    @property
    def grammar_rule_type(self):
        """The GrammarRule's type."""
        return self.function_type

    @property
    def grammar_rule_desc(self):
        """A string containing a description of the associated grammar rule."""
        return self.rule.description


    def __call__(self, args):
        """Convert the list of arguments `args` into a dictinonary.

        All keywords are removed from `args`.
        All variable names become keys to their corresponding values.
        VarLists are converted into lists of tuples, where each tuples
        contains the values of the variables contained in the list in order.

        Args:
            args (list): A list of strings

        Returns:
            (dict) A dictionary, which maps from variable names to values
        """
        return self.rule(args)


    @classmethod
    def _parse_grammar_desc(cls, input_str):
        """Parse a string containg a grammar description.

        Args:
            input_str (str): A string containing grammar rules.

        Returns:
            (tuple) list of lines with grammar rules and list with choices
        """
        lines = [line.strip() for line in input_str.splitlines()]
        idx = lines.index("Grammar:")

        grammar_lines = []
        for line in lines[idx+1:]:
            idx += 1 # keep track of line number for "choices from a list"
            grammar_lines.append(line.strip("\\ \t\n"))
            # grammar rules may span multiple lines (via "\" at the end)
            if not line.endswith("\\"):
                break

        choice_lines = []
        for line in lines[idx+1:]:
            # skip empty lines
            if not line.strip():
                continue
            # A "choice from a list" line
            if ":=" in line and "{" in line and "}" in line:
                choice_lines.append(line)
            else:
                break

        try:
            function_type = lines[lines.index("Type:") + 1].strip()
        except ValueError:
            function_type = None

        return grammar_lines, choice_lines, function_type


    @classmethod
    def _extract_rule_name(cls, grammar_desc):
        """Extract the name of the rule from the grammar_desc.

        Args:
            grammar_desc (list): List of strings as return by _prepare_grammar_desc

        Returns:
            (tuple) The name of the grammar rule and the updated grammar_desc
        """
        if "(" in grammar_desc[0]:
            stop_idx = [")" in t for t in grammar_desc].index(True) + 1
        else:
            matches = [("$" in t or "!" in t) for t in grammar_desc]
            stop_idx = matches.index(True) if True in matches else len(grammar_desc)

        return " ".join(grammar_desc[0:stop_idx]).strip(" )("), grammar_desc[stop_idx:]


    @classmethod
    def _extract_instructions(cls, grammar_desc):
        """Extract the instruction on how to parse the grammar from a list of strings.

        Args:
            grammar_lines (list): A list of string with grammar rules

        Returns:
            (list) The parsing instructions.
        """
        instructions = []
        for token in grammar_desc:
            if token.startswith("!"):
                instructions.append(Expression(token[1:]))
            elif not token.startswith("$"):
                instructions.append(Keyword(token))
            elif "[" in token and token.endswith("]"):
                bracket_idx = token.index("[")
                name = token[1:bracket_idx]
                data = cls._extract_instructions(token[bracket_idx+1:-1].split())
                instructions.append(VarList(name, data))
            else:
                instructions.append(Variable(token[1:]))
        return instructions


    @classmethod
    def _extract_choices(cls, choice_lines):
        """Exract the choices from a list of strings.

        Args:
            grammar_lines (list): A list of string with choice rules.

        Returns:
            (dict) A list of choices for any variable if any.
        """
        choices = {}
        for line in choice_lines:
            var, vals = line.split(":=")
            choices[var.strip()] = [val.strip(" }{") for val in vals.split(",")]
        return choices


    @classmethod
    def _prepare_grammar_desc(cls, grammar_lines):
        """Prepares the a set of string containing grammar rules for futher processing.

        Args:
            grammar_lines (list): A list of strings containing grammar rules.

        Returns:
            (list) A processed list of strings containing grammar rules.

        Raises:
            ValueError: If the number of opening and closing square brackets do not match.
        """
        grammar_desc = " ".join(grammar_lines).split(" ")
        open_idx = [idx for idx, token in enumerate(grammar_desc) if "[" in token]
        close_idx = [idx+1 for idx, token in enumerate(grammar_desc) if "]" in token]
        if len(open_idx) != len(close_idx):
            raise ValueError("Number of '[' does not match the number of ']'")

        for o_idx, c_idx in zip(reversed(open_idx), reversed(close_idx)):
            grammar_desc[o_idx:c_idx] = [" ".join(grammar_desc[o_idx:c_idx])]
        return grammar_desc




class GrammarFunction:
    """A GrammarFunction is wrapper around a function describing a grammar rule.

    It uses a GrammarRule internally to convert the `args` list of the function
    into the format returned by the Grammar Rule.
    If the `grammar_desc` parameter is ommitted, the docstring of `fun` will
    be used instead.

    Note:
        This docstring differs for every instance of GrammarFunction,
        that is the __init__ methods modifies the docstring.

    Args:
        fun (function): The function to be wrapped
        grammar_desc (str): A string describing the grammar.
        expr_rule (ExpressionRule): The type of expression rule to be used
    """

    def __init__(self, fun, grammar_desc=None, expr_rule=ExpressionRule):
        """Please see help(GrammarFunction)."""
        grammar_desc = dedent(grammar_desc) if not grammar_desc is None else fun.__doc__
        self._grammar_parser = GrammarParser(grammar_desc, expr_rule)
        self._fun = convert_function(fun)
        # override the docstring of this instance operator
        self.__name__ = fun.__name__
        self.__doc__ = grammar_desc


    @property
    def grammar_parser(self):
        """The Parser used for this grammar."""
        return self._grammar_parser

    @property
    def grammar_rule(self):
        """The GrammarRule, responsible for argument conversion."""
        return self._grammar_parser.rule

    @property
    def grammar_rule_name(self):
        """The GrammarRule's name."""
        return self._grammar_parser.name

    @property
    def grammar_rule_type(self):
        """The GrammarRule's type."""
        return self._grammar_parser.grammar_rule_type

    @property
    def grammar_rule_desc(self):
        """A string containing a description of the associated grammar rule."""
        return self._grammar_parser.grammar_rule_desc

    @property
    def function(self):
        """The function this GrammarRule wrappes around."""
        return self._fun


    def __call__(self, code, args, env=None):
        """Apply the internal `function` to code and args.

        The `args` argument is parsed by the internal `GrammarParser`
        before passing it to the `function`.

        Args:
            code (str): The already generated code
            args (list): A list of string representing DSL arguments
            env (dict): A dictionary with environment variables.
        Returns:
            (str) The modified code.

        Raises:
            DSLFunctionError: If a DSLArgumentError occurs while applying the grammar rules.
        """
        try:
            return self._fun(code, self._grammar_parser(args), env)
        except DSLArgumentError as err:
            raise DSLFunctionError("Missing argument", err.rule, self.__doc__) from err




@decorator_factory
def grammar(fun, doc=None, expr=ExpressionRule):
    """Decorator for external grammar rules.

    If no grammar is specified in the functions docstring
    or via the `doc` argument this will raise an error.
    Otherwise the argument list of the function will be converted into a dictonary,
    which maps grammar variables names to values.

    Args:
        fun (function): The function to be wrapped
        doc (str): A string containing the grammar description
        expr (ExpressionRule): The type of expression rule to be used

    Returns:
        (function) GrammarFunction - functor, which behaves similar to `fun`
    """
    return GrammarFunction(fun, doc, expr)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""The DSL provides a mechanism to add new grammar rules via DSL statements.

This mechanism is implemented by the PipeFunction class, which synthesizes
new grammar rules based on a DSL statement. The Converter classes translate
the arguments of the new grammar.
"""

from .types import is_variable, is_expression, is_var_list
from .rules import GrammarRule
from .exceptions import DSLArgumentError, DSLFunctionError




class IdentityConverter:
    """A converter for keyword and value arguments.
    It basically does nothing, it just exists for API consistency.

    Args:
        arg (dict): A dict containg a Keyword or Value (see nldsl.core.types).
        name_dict (dict): A mapping from variable names to instructions.
    """

    def __init__(self, arg, name_dict):
        """Please see help(IdentityConverter)."""
        self._value = arg.data


    @property
    def value(self):
        """The value (or keyword) which will be returned by the call method"""
        return self._value


    def __call__(self, input_dict):
        """Returns a list containing only the value of the `arg` element
        it received on constructions.

        Args:
            input_dict (dict): Not used, exist only for API consistency.

        Returns:
            (list) A list containing only a single string.
        """
        return [self._value]




class NameConverter:
    """Converts a variable or expression name to its values.

    Args:
        arg (dict): Either a Variable or Expression type (see nldsl.core.types).
        name_dict (dict): A mapping from variable names to instructions.

    Raises:
        ValueError: If the arg.data is not a key in `name_dict`.
    """

    def __init__(self, arg, name_dict):
        """Please see help(NameConverter)."""
        if not arg.data in name_dict:
            raise ValueError("Variable name not found.")
        self._name = arg.data


    @property
    def name(self):
        """The name of the variable or expression."""
        return self._name


    def __call__(self, input_dict):
        """Returns the value of the variable name stored in `input_dict`.

        Args:
           input_dict (dict): A map from variable names to values

        Returns:
            (list) A list containing only one string
        """
        return [input_dict[self._name]]




class ListConverter:
    """Converts a variable list to its values.

    Args:
        arg (dict): A dict containg a VarList (see nldsl.core.types).
        name_dict (dict): A mapping from variable names to instructions.

    Raises:
        ValueError: If the variable list name is not in `name_dict`.
    """

    def __init__(self, arg, name_dict):
        """Please see help(ListConverter)."""
        self._name = arg.name
        if not self._name:
            variables = {x.data for x in arg.data if is_variable(x)}
            self._name = next(name for name in name_dict if set(name) == variables)
        elif not self._name in name_dict:
            raise ValueError("Variable list name not found.")

        input_vars = [v.data for v in name_dict[self._name] if is_variable(v)]
        self._map = {x.data: input_vars.index(x.data) for x in arg.data if is_variable(x)}
        self._data = arg.data


    @property
    def name(self):
        """The name of the variable list."""
        return self._name

    @property
    def data(self):
        """The data of the target."""
        return self._data

    @property
    def map(self):
        """The mapping of variable positions."""
        return self._map


    def __call__(self, input_dict):
        """Returns the values for the variable list stored in `input_dict`.

        Args:
           input_dict (dict): A map from variable names to values

        Returns:
            (list) A list containg the values corresponding to variable list
        """
        return sum([self._map_input(data) for data in input_dict[self._name]], [])


    def _map_input(self, input_data):
        """Map a tuple of values to a list.

        This list contains all keywords and the values of variables
        found in the tuple in the correct order,
        as defined by the variable list.

        Args:
            input_data (tuple): A tuple of values

        Returns:
            (list) A list of strings.
        """
        res = []
        for arg in self._data:
            if is_variable(arg):
                res.append(input_data[self._map[arg.data]])
            else:
                res.append(arg.data)
        return res




class PipeFunction:
    """PipeFunction provides a mechanism to define functions within the DSL.

    Given a DSL statement such as::

        #$ prepare $x with labels $y = rename columns $x to data, $y to labels | sort by $y

    The left side of the equal sign defines the `instructions` and from the right
    side we derive the `fun_map`, which maps functions to partially specified argument lists.

    Note:
        This docstring differs for every instance of GrammarFunction,
        that is the __init__ methods modifies the docstring.

    Args:
        name (str): The name of this PipeFunction within the DSL.
        instructions (list): A list of argument parsing instructions.
        fun_list (list): An list of (function, argument list) pairs.
    """

    def __init__(self, name, instructions, fun_list):
        """Please see help(PipeFunction)."""
        self.name = name
        self.rule = GrammarRule(instructions, {})
        self.fun_conv = self._extract_args_converters(instructions, fun_list)

        self.function_type = self._compute_type(fun_list)
        self.__name__ = self.name.replace(" ", "_")
        self.__doc__ = self._build_docstring()


    @property
    def grammar_rule(self):
        """The GrammarRule, responsible for argument conversion."""
        return self.rule

    @property
    def grammar_rule_name(self):
        """The GrammarRule's name."""
        return self.name

    @property
    def grammar_rule_desc(self):
        """A string containing a description of the associated grammar rule."""
        return self.rule.description

    @property
    def grammar_rule_type(self):
        """The GrammarRule's type."""
        return self.function_type

    @property
    def functions(self):
        """A list of the internal functions."""
        return [fun for fun, conv in self.fun_conv]

    @property
    def argument_converters(self):
        """A list of (function, argument converters) pairs."""
        return self.fun_conv


    def __call__(self, code, args, env=None):
        """Map the arguments `args` into a DSL pipeline and convert it to code.

        Args:
            args (list): A list of strings.

        Returns:
            (str) A string containing executable python code.

        Raises:
            DSLFunctionError: If a DSLArgumentError occurs while applying the grammar rules.
        """
        try:
            args_dict = self.rule(args)
        except DSLArgumentError as err:
            raise DSLFunctionError("Missing argument", err.rule, self.__doc__) from err

        for fun, arg_converters in self.fun_conv:
            code = fun(code, sum([conv(args_dict) for conv in arg_converters], []), env)
        return code


    @classmethod
    def _extract_name_dict(cls, instructions):
        """Extract the names of variables and variable list from the `instructions`.

        Args:
            instructions (list): A list of parsing instructions.

        Returns:
            (dict) A mapping from variable names to (modified) instructions.
        """
        name_dict = {}
        for i in instructions:
            if is_variable(i) or is_expression(i):
                name_dict[i.data] = i.data
            elif is_var_list(i) and i.name:
                name_dict[i.name] = i.data
            elif is_var_list(i) and not i.name:
                vl_name = tuple([x.data for x in i.data if is_variable(x)])
                name_dict[vl_name] = i.data
        return name_dict


    def _extract_args_converters(self, instructions, fun_list):
        """Extract the necessary argument converters from `instructions`
        and `fun_list` and create them.

        Args:
            instructions (list): A list of parsing instructions
            fun_list (list): An list of (function, argument list) pairs.

        Returns:
            (list) A list of argument converters.
        """
        fun_conv, name_dict = [], self._extract_name_dict(instructions)
        for fun, args in fun_list:
            conv = []
            for arg in args:
                if is_var_list(arg):
                    conv.append(ListConverter(arg, name_dict))
                elif is_variable(arg):
                    conv.append(NameConverter(arg, name_dict))
                elif is_expression(arg):
                    conv.append(NameConverter(arg, name_dict))
                else:
                    conv.append(IdentityConverter(arg, name_dict))
            fun_conv.append((fun, conv))
        return fun_conv


    @classmethod
    def _compute_type(cls, fun_list):
        type_list = []
        for fun, _ in fun_list:
            if hasattr(fun, "grammar_rule_type"):
                type_list.append(fun.grammar_rule_type)
            else:
                type_list.append(None)

        if any(t is None for t in type_list):
            return None
        if any(t == "Function" for t in type_list):
            return "Function"
        if all(t == "Transformation" for t in type_list):
            return "Transformation"
        if type_list[0] == "Initialization":
            return "Initialization"
        if type_list[-1] == "Operation":
            return "Operation"
        return None


    def _build_docstring(self):
        # placeholder
        return ""



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""The Pandas extension provides the code generator, which targes Pandas.

Examples:
    Create an instance of a SparkCodeGenerator::

        from nldsl import SparkCodeGenerator

        self.code_gen = SparkCodeGenerator()

    Generated code: x = y.filter(y.col1 == m & y.col3.isin([v1, v2, v3]))::

        model = "## x = on y | select rows y.col1 == m and y.col3 in [v1, v2, v3]"
        result = self.code_gen(model)

    Generated code: x = df.dropDuplicates()::

        model = "## x = on df | drop duplicates"
        result = self.code_gen(model)

    DSL statement that adds new rules to the grammer::

        model_def = #$ sort and clean $x = sort by $x | drop duplicates
        model_eval = ## x = on df | sort and clean col1
        result = self.code_gen(model_def + mode_eval)
"""

from nldsl.core import CodeGenerator, ExpressionRule, OperatorType, grammar, list_to_string
from nldsl import docs




class SparkCodeGenerator(CodeGenerator):
    """ A SparkCodeGenerator translates DSL statements into executable spark code.

    There are two kind of DSL statements, the ones which can be evaluate to executable code
    and the ones, which extend the DSL Grammer.
    As a result parsing a set of DSL statements usually has two impacts.
    Executable code is generate and the SparkCodeGenerator modifies itself in such a way
    that he is capable of parsing statements according to the new Grammer rules.

    Furthermore the SparkCodeGenerator derives from the CodeMap class and his grammer can also
    be extended by rules, which can not be expressed within the DSL.
    This is done with the __setitem__ and registerFunction methods.

    Example:
        Adding new rules::

            def show(code, args):
                return "print({})".format(code)

        SparkCodeGenerator.registerFunction("show entire table", show) # Add to class
        myCodeGen = SparkCodeGenerator();
        myCodeGen["show entire table"] = show # only add to this instance

    Args:
        recommend (bool): Whether to return a Recommendation if possible or always raise an error.
        spark_name (str): The name of the variable which holds/should hold the spark session
        import_name (str): The name under which the spark session is imported.
        start_session_named (str): If not empty a SparkSession with this name will be created.
        stop_session (bool): If True code will be added to stop the SparkSession
        kwargs (dict): Additional keyword argument, which will be added to the environment
    """

    def __init__(self, recommend=True, spark_name="spark", import_name="SparkSession",
                 start_session_named="", stop_session=False, **kwargs):
        """Please see help(SparkCodeGenerator)."""
        super().__init__(recommend=recommend, spark_name=spark_name,
                         import_name=import_name, **kwargs)
        self.start_session_named = start_session_named
        self.stop_session = stop_session


    def postprocessing(self, code_list):
        """ Optionally adds lines of code, which create and stop a SparkSession.
        The behavior of this function is determined by the `start_session_named`
        and `stop_session` parameters of the __init__ method.

        Args:
            code_list (list): A list of string containing executable code.
        Returns:
            (list) The modified code list
        """
        if self.start_session_named:
            spark_str = self.env["spark_name"] + " = " + self.env["import_name"]
            spark_str += ".builder.appName('{}').getOrCreate()".format(self.start_session_named)
            code_list = [spark_str] + code_list
        if self.stop_session:
            code_list.append(self.env["spark_name"] + ".stop()")
        return code_list


    def extract_environment(self, source_lines):
        """Extract the name of the spark session from `source_lines` if it exists.

        Args:
            source_lines (list): A list of strings containing source code.
        """
        imps = []
        for line in source_lines:
            sline = line.strip()
            if ".builder." in line and ".getOrCreate()" in line and not sline.startswith("#"):
                imps.append(sline)
        if imps:
            self.env["spark_name"] = imps[-1].split("=")[0].strip()




class SparkExpressionRule(ExpressionRule):
    """An ExpressionRule dedicated to parsing Spark expressions.

    Args:
        expr_name (str): The name of the expression.
        next_keyword (str): The keyword following the expression or None
    """

    def __init__(self, expr_name, next_keyword):
        super().__init__(expr_name, next_keyword)
        self.operator_map["and"] = " & "
        self.operator_map["or"] = " | "
        self.operator_map["in"] = ".isin"
        self.operator_map["not"] = "~"

        self.operator_type["in"] = OperatorType.UNARY_FUNCTION




@grammar(docs.EXPRESSION_ONLY_DOC)
def expression_only(code, args):
    return code + args["expr"]


@grammar(docs.ON_DF_DOC)
def on_dataframe(code, args):
    return code + args["dataframe"]


@grammar(docs.CREATE_DF_DOC)
def create_dataframe(code, args, env):
    cols = list_to_string(args["header"])
    return code + env["spark_name"] + ".createDataFrame({}, schema={})".format(args["data"], cols)


@grammar(docs.LOAD_FROM_DOC)
def load_from(code, args, env):
    return code + "{}.read.format('{}').load({})".format(env["spark_name"], args["type"], args["path"])


@grammar(docs.SAVE_TO_DOC)
def save_to(code, args):
    return code + ".write.format('{}').save({})".format(args["type"], args["path"])


@grammar(docs.UNION_DOC)
def union(code, args):
    return code + ".unionByName({})".format(args["table"])


@grammar(docs.DIFFERENCE_DOC)
def difference(code, args):
    return code + ".subtract({})".format(args["table"])


@grammar(docs.INTERSECTION_DOC)
def intersection(code, args):
    return code + ".intersect({})".format(args["table"])


@grammar(docs.SELECT_COLUMNS_DOC)
def select_columns(code, args):
    return code + ".select({})".format(list_to_string(args["columns"]))


@grammar(docs.SELECT_ROWS_DOC, SparkExpressionRule)
def select_rows(code, args):
    return code + ".filter({})".format(args["condition"])


@grammar(docs.DROP_COLUMNS_DOC)
def drop_columns(code, args):
    return code + ".drop({})".format(list_to_string(args["columns"]))


@grammar(docs.JOIN_DOC)
def join(code, args):
    cols = list_to_string(args["columns"])
    return code + ".join({}, on={}, how='{}')".format(args["table"], cols, args["how"])


@grammar(docs.GROUP_BY_DOC)
def group_by(code, args):
    cols = list_to_string(args["columns"])
    return code + ".groupBy({}).{}()".format(cols, args["aggregation"])


@grammar(docs.REPLACE_VALUES_DOC)
def replace_values(code, args):
    return code + ".replace({}, {})".format(args["old_value"], args["new_value"])


@grammar(docs.APPEND_COLUMN_DOC)
def append_column(code, args):
    cols_str = ".withColumn({})"
    return code + "".join(cols_str.format(", ".join([args["col_name"], args["col_expr"]])))


@grammar(docs.SORT_BY_DOC)
def sort_by(code, args):
    cols, order = zip(*args["columns"])
    cols = list_to_string(cols)
    order = [o == "ascending" for o in order]
    return code + ".sort({}, ascending={})".format(cols, order)


@grammar(docs.DROP_DUPLICATES_DOC)
def drop_duplicates(code):
    return code + ".dropDuplicates()"


@grammar(docs.RENAME_COLUMNS_DOC)
def rename_columns(code, args):
    ops = ["withColumnRenamed({}, {})".format(o, n) for o, n in args["columns"]]
    return code + "." + ".".join(ops)


@grammar(docs.SHOW_DOC)
def show(code):
    return code + ".show()"


@grammar(docs.SHOW_SCHEMA_DOC)
def show_schema(code):
    return code + ".printSchema()"


@grammar(docs.DESCRIBE_DOC)
def describe(code):
    return code + ".describe()"


@grammar(docs.HEAD_DOC)
def head(code, args):
    return code + ".head({})".format(args["num_rows"])


@grammar(docs.COUNT_DOC)
def count(code):
    return code + ".count()"



SparkCodeGenerator.register_function(expression_only)
SparkCodeGenerator.register_function(on_dataframe)
SparkCodeGenerator.register_function(create_dataframe)
SparkCodeGenerator.register_function(load_from)
SparkCodeGenerator.register_function(save_to)
SparkCodeGenerator.register_function(union)
SparkCodeGenerator.register_function(difference)
SparkCodeGenerator.register_function(intersection)
SparkCodeGenerator.register_function(select_columns)
SparkCodeGenerator.register_function(select_rows)
SparkCodeGenerator.register_function(drop_columns)
SparkCodeGenerator.register_function(join)
SparkCodeGenerator.register_function(group_by)
SparkCodeGenerator.register_function(replace_values)
SparkCodeGenerator.register_function(append_column)
SparkCodeGenerator.register_function(sort_by)
SparkCodeGenerator.register_function(drop_duplicates)
SparkCodeGenerator.register_function(rename_columns)
SparkCodeGenerator.register_function(show)
SparkCodeGenerator.register_function(show_schema)
SparkCodeGenerator.register_function(describe)
SparkCodeGenerator.register_function(head)
SparkCodeGenerator.register_function(count)



#! python3

# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

import sys
import argparse
import re
from enum import Enum
from textx.exceptions import TextXSyntaxError
from nldsl import PandasCodeGenerator, SparkCodeGenerator, Recommendation




class StmtConfig(Enum):
    EVAL_PREFIX = "##"
    DEFINE_PREFIX = "#$"
    SHEBANG_PREFIX = "#!"
    IMPORT_STRING = "import"


class Stmt:
    def __init__(self, value, indent):
        self.value = value
        self.indent = indent


class PythonStmt(Stmt): pass
class CodeStmt(PythonStmt): pass
class ImportStmt(PythonStmt): pass
class ShebangStmt(PythonStmt): pass


class DSLStmt(Stmt): pass
class DefineStmt(DSLStmt): pass
class EvalStmt(DSLStmt): pass


def make_stmt(value, indent):
    if value.startswith(StmtConfig.EVAL_PREFIX.value):
        return EvalStmt(value, indent)

    if value.startswith(StmtConfig.DEFINE_PREFIX.value):
        return DefineStmt(value, indent)

    if value.startswith(StmtConfig.SHEBANG_PREFIX.value):
        return ShebangStmt(value, indent)

    if StmtConfig.IMPORT_STRING.value in value:
        return ImportStmt(value, indent)

    return CodeStmt(value, indent)


def is_python(stmt):
    return isinstance(stmt, PythonStmt)

def is_code(stmt):
    return isinstance(stmt, CodeStmt)

def is_import(stmt):
    return isinstance(stmt, ImportStmt)

def is_shebang(stmt):
    return isinstance(stmt, ShebangStmt)

def is_dsl(stmt):
    return isinstance(stmt, DSLStmt)

def is_define(stmt):
    return isinstance(stmt, DefineStmt)

def is_eval(stmt):
    return isinstance(stmt, EvalStmt)

    


class NLDSLCompiler:
    TARGET_PATTERN = re.compile(r"\s?##\starget_code\s=")

    def __init__(self, code_gen, auto_import):
        self.code_gen = code_gen
        self.imports = auto_import


    def compile(self, input_file, output_file):
        with open(input_file, "r") as code:
            stmts = []
            for idx, line in enumerate(code):
                # skip target annotations
                if self.TARGET_PATTERN.match(line):
                    continue
                sline = line.lstrip().rstrip("\n")
                stmts.append(make_stmt(sline, len(line) - len(sline) - 1))

            try:
                dsl_code = self.code_gen("\n".join([s.value for s in stmts if is_dsl(s)]))
            except TextXSyntaxError:
                print("ERROR: No NLDSL Code found --- aborting compilation.", file=sys.stderr)
                sys.exit(1)

        result = self._compile(self._add_imports(stmts), dsl_code)

        with open(output_file, "w") as code:
            code.write("\n".join(result))
    

    @classmethod
    def substitute_code(cls, stmt, dsl_code):
        if is_eval(stmt):
            dsl_stmt = dsl_code.pop(0)
            if isinstance(dsl_stmt, Recommendation):
                raise SyntaxError("Invalid DSL statement.")
            return " " * stmt.indent + dsl_stmt
        return " " * stmt.indent + stmt.value if stmt.value else ""


    def _add_imports(self, stmts): 
        idx, indent = next(((i, s.indent) for i, s in enumerate(stmts) if is_import(s)), (0, 0))
        idx = idx + 1 if idx == 0 and is_shebang(stmts[0]) else idx
        return stmts[0:idx] + [ImportStmt(imp, indent) for imp in self.imports] + stmts[idx:]




class PandasCompiler(NLDSLCompiler):
    def __init__(self, auto_import=False, **kwargs):
        prefix = ""
        if "import_name" in kwargs and kwargs["import_name"] != "pandas":
            prefix = " as " + kwargs["import_name"]
 
        auto_import = ["import pandas" + prefix] if auto_import else []
        super().__init__(PandasCodeGenerator(**kwargs), auto_import)


    def _compile(self, stmts, dsl_code):
        return [self.substitute_code(stmt, dsl_code) for stmt in stmts if not is_define(stmt)]




class SparkCompiler(NLDSLCompiler):
    def __init__(self, auto_import=False, **kwargs):
        prefix = ""
        if "import_name" in kwargs and kwargs["import_name"] != "SparkSession":
            prefix = " as " + kwargs["import_name"]

        auto_import = ["from pyspark.sql import SparkSession" + prefix] if auto_import else []
        super().__init__(SparkCodeGenerator(**kwargs), auto_import)


    def _compile(self, stmts, dsl_code):
        if self.code_gen.start_session_named:
            stmts = self._insert_start_stmt(stmts)
        if self.code_gen.stop_session:
            stmts = self._insert_stop_stmt(stmts)
        return [self.substitute_code(stmt, dsl_code) for stmt in stmts if not is_define(stmt)]


    def _insert_start_stmt(self, stmts):
        idx, indent = next(((i, s.indent) for i, s in enumerate(stmts) if is_dsl(s)), (0, 0))
        stmts.insert(idx, EvalStmt("", indent))
        stmts.insert(idx + 1, CodeStmt("", indent))
        return stmts
         

    def _insert_stop_stmt(self, stmts):
        idx, indent = next(((i, s.indent) for i, s in enumerate(stmts[::-1]) if is_dsl(s)), (0, 0))
        stmts.insert(len(stmts) - idx, EvalStmt("", indent))
        stmts.insert(len(stmts) - idx - 1, CodeStmt("", indent))
        return stmts




if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Compile NLDSL code to python code.")

    parser.add_argument("input_file", help="A python file containing DSL code.")
    parser.add_argument("output_file", help="The file in which the compiled code shall be stored.")

    parser.add_argument("--target", "-t", dest="target", required=True,
                        help="The target framework, must be either pandas or spark).")
    parser.add_argument("--start-session", "-s", dest="start", default="",
                        help="(spark only) If provided this will start a session under the given name.")
    parser.add_argument("--stop-session", "-o", action="store_true", dest="stop",
                        help="(spark only) If provide a stop session statement will be added.")
    parser.add_argument("--spark-variable", "-v", dest="spark_variable", default="spark",
                        help="(spark only) The name of the variable which holds/should hold the session.")
    parser.add_argument("--auto-import", "-a", action="store_true", dest="auto_import",
                        help="Whether or not import statements shall be added if necessary")
    parser.add_argument("--import-name", "-n", dest="import_name", default="infer",
                        help="The name under which the target framework is/should be imported")

    args = parser.parse_args()
   
    if args.target == "spark":
        args.import_name = "SparkSession" if args.import_name == "infer" else args.import_name
        spark_args = {
            "start_session_named": args.start,
            "stop_session": args.stop,
            "import_name": args.import_name,
            "spark_name": args.spark_variable
        }
        compiler = SparkCompiler(args.auto_import, **spark_args)

    elif args.target == "pandas":
        args.import_name = "pandas" if args.import_name == "infer" else args.import_name
        pandas_args = {"import_name": args.import_name}
        compiler = PandasCompiler(args.auto_import, **pandas_args)

    compiler.compile(args.input_file, args.output_file)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""Each argument provided to a DSL statement is of a certain type.
The DSL argument types are Value, Keyword, Variable, Expression and VarList.
Additionally some types dedicated to the operators used by expression exists.
"""

from enum import IntEnum




class OperatorType(IntEnum):
    """Type of operators as used by the ExpressionRule."""
    OPERATOR = 0
    UNARY_FUNCTION = 1
    BINARY_FUNCTION = 2




class Type:
    """Base class for DSL argument types.

    Args:
        data (object): The data associated with this object
    """

    def __init__(self, data):
        """Please see help(Type)."""
        self._data = data

    @property
    def data(self):
        """Get the data member."""
        return self._data

    @data.setter
    def data(self, new_data):
        """Set the data member."""
        self._data = new_data
        return self




class Value(Type):
    """A Type for arguments which are plain values."""

    @property
    def value(self):
        """Set the value."""
        return self.data

    @value.setter
    def value(self, new_value):
        """Get the value."""
        self.data = new_value
        return self




class Keyword(Type):
    """A Type for arguments which are keywords."""

    @property
    def keyword(self):
        """Get the keyword."""
        return self.data

    @keyword.setter
    def keyword(self, new_keyword):
        """Set the keyword."""
        self.data = new_keyword
        return self


class Variable(Type):
    """A Type for arguments which are variables."""

    @property
    def name(self):

        """Get the name."""
        return self.data

    @name.setter
    def name(self, new_name):
        """Set the name."""
        self.data = new_name
        return self


class Expression(Type):
    """A Type for arguments which are expressions.

    Args:
        name (str): The name of this variable
    """

    def __init__(self, name):
        """Please see help(Expression)."""
        super().__init__(name)
        self.name = name

    @property
    def name(self):
        """Get the name."""
        return self.data

    @name.setter
    def name(self, new_name):
        """Set the name."""
        self.data = new_name
        return self


class VarList(Type):
    """A Type for arguments which are Variable Lists.

    Args:
        name (str): The name of this variable list
        variables (list): A list of Variable or Keyword objects
    """

    def __init__(self, name, variables):
        """Please see help(VarList)."""
        super().__init__(variables)
        self._name = name

    @property
    def name(self):
        """Get the name."""
        return self._name

    @name.setter
    def name(self, new_name):
        """Set the name."""
        self._name = new_name
        return self

    @property
    def variables(self):
        """Get the variables and keywords."""
        return self.data

    @variables.setter
    def variables(self, new_variables):
        """Set the variables and keywords."""
        self.data = new_variables
        return self




def is_type(arg):
    """Check if `arg` is an instance of Type.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a Type, false otherwise.
    """
    return isinstance(arg, Type)




def is_value(arg):
    """Check if `arg` is an instance of Value.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a Value, false otherwise.
    """
    return isinstance(arg, Value)




def is_keyword(arg):
    """Check if `arg` is an instance of Keyword.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a Keyoword, false otherwise.
    """
    return isinstance(arg, Keyword)




def is_variable(arg):
    """Check if `arg` is an instance of Variable.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a Variable, false otherwise.
    """
    return isinstance(arg, Variable)




def is_expression(arg):
    """Check if `arg` is an instance of Expression.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a Expression, false otherwise.
    """
    return isinstance(arg, Expression)




def is_var_list(arg):
    """Check if `arg` is an instance of VarList.

    Args:
        arg (object): The object to investigate.

    Returns:
        (bool) True if the argument is a VarList, false otherwise.
    """
    return isinstance(arg, VarList)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""The core package implements every aspect of nldsl that is not target dependent.

Its primary interface is the CodeGenerator class, which serves as a common base
class for every CodeGenerator with a specific target. The CodeGenerator also
implements the dictonary API via inheritance from the CodeMap.

Adding new grammar rules via python is implemented by the GrammarFunction
and its associated GrammarParser. The grammar decorator provides syntactic sugar
for generating instances of GrammarFunction from normal python functions.

The PipeFunction is called into action once a DSL statement, which defines a new
grammar rule, is observed by the CodeGenerator. The PipeFunction then synthesizes a
new pipeline using the already exist dynamic grammar rules.
"""

from .types import (OperatorType, Type, Value, Keyword, Variable, Expression, VarList,
                    is_type, is_value, is_keyword, is_variable, is_expression, is_var_list)
from .utils import decorator_factory, cast_down, convert_function, split_code, list_to_string
from .rules import KeywordRule, VariableRule, VarListRule, ExpressionRule, GrammarRule
from .external import GrammarParser, GrammarFunction, grammar
from .internal import IdentityConverter, NameConverter, ListConverter, PipeFunction
from .exceptions import DSLError, DSLArgumentError, DSLFunctionError
from .codegen import CodeMap, CodeGenerator, Recommendation



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""A selection of  exception classes to describe incomplete or incorrect DSL Arguments.
All of these exceptions will be catched by the CodeGenerator class and converted into
a code recommendation.
"""




class DSLError(Exception):
    """Base class for all DSL related exceptions."""




class DSLArgumentError(DSLError):
    """This exception indicates that a DSL argument is missing or incorrect.

    Args:
        message (str): A string containing a description of the error
        rule (grammar rule): An instance of grammar rule class.
    """

    def __init__(self, message, rule):
        """Please see help(DSLArgumentError)."""
        super().__init__("message: {}, rule: {}".format(message, rule))
        self.message = message
        self.rule = rule


    def __reduce__(self):
        """Allow this exceptions to be pickled.

        Returns:
            (tuple) DSLArgumentError and a tuple of the constructor arguments.
        """
        return (DSLArgumentError, (self.message, self.rule))




class DSLFunctionError(DSLError):
    """This exception indicates that a DSL function failed to generate code

    This exceptions is usually raised due to a previous DSLArgumentError exception.

    Args:
        message (str): A string containing a description of the error
        rule (grammar rule): An instance of grammar rule class.
    """

    def __init__(self, message, rule, doc=""):
        """Please see help(DSLFunctionError)."""
        super().__init__("message: {}, rule: {}".format(message, rule))
        self.message = message
        self.rule = rule
        self.doc = doc


    def __reduce__(self):
        """Allow this exceptions to be pickled.

        Returns:
            (tuple) DSLFunctionError and a tuple of the constructor arguments.
        """
        return (DSLFunctionError, (self.message, self.rule))



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""The Pandas extension provides the code generator, which targes Pandas.

Examples:
    Create an instance of a PandasCodeGenerator::

        from nldsl import PandasCodeGenerator()
        code_gen = PandasCodeGenerator()

    Generate the code: x = y[y.col1 == m & y.col3.isin([v1, v2, v3])]::

        model = "## x = on y | select rows y.col1 == m and y.col3 in [v1, v2, v3]"
        result = self.code_gen(model)[0]

    Generate the code: print(df.rename(columns={'old1': 'new1', 'old2': 'new2'}))::

        model = "## on df | rename columns old1 to new1, old2 to new2 | show"
        result = self.code_gen(model)[0]

    Define a new rule and use it to generate code::

        model_def = "#$ my pipeline $[$old to $new] = rename columns $[$old to $new] | show"
        model_eva = "## on df | my pipeline old1 from new1, old2 from new2"
        result = self.code_gen(model_def + model_eva)[0]
"""

import re
from nldsl.core import (CodeGenerator, ExpressionRule, OperatorType, grammar,
                        split_code, list_to_string)
from nldsl import docs




class PandasCodeGenerator(CodeGenerator):
    """ A PandasCodeGenerator translates DSL statements into executable pandas code.

    There are two kind of DSL statements, the ones which can be evaluate to executable code
    and the ones, which extend the DSL Grammer.
    As a result parsing a set of DSL statements usually has two impacts.
    Executable code is generate and the PandasCodeGenerator modifies itself in such a way
    that he is capable of parsing statements according to the new Grammer rules.

    Furthermore the PandasCodeGenerator derives from the CodeMap class and his grammer can also
    be extended by rules, which can not be expressed within the DSL.
    This is done with the __setitem__ and registerFunction methods.

    Example:
        Adding new rules::

            def show(code, args):
                return "print({})".format(code)

            PandasCodeGenerator.registerFunction("show entire table", show) # Add to class

            myCodeGen = PandasCodeGenerator();
            myCodeGen["show entire table"] = show # only add to this instance

    Args:
        recommend (bool): Whether to return a Recommendation if possible or always raise an error.
        import_name (str): The name under which the Pandas module is imported
        kwargs (dict): Additional keyword argument, which will be added to the environment
    """
    pandas_import_regex = re.compile(r"^\s*import\s*pandas\s*as\s*[a-zA-Z_][a-zA-Z_0-9]*\s*$")


    def __init__(self, recommend=True, import_name="pandas", **kwargs):
        """Please see help(PandasCodeGenrator)."""
        super().__init__(recommend=recommend, import_name=import_name, **kwargs)


    def extract_environment(self, source_lines):
        """Extract the name under which pandas is import from `source_lines` if it exists.

        Args:
            source_lines (list): A list of strings containing source code.
        """
        imps = [l for l in source_lines if self.pandas_import_regex.match(l)]
        if imps:
            self.env["import_name"] = imps[-1].split(" as ")[-1].strip()




class PandasExpressionRule(ExpressionRule):
    """An ExpressionRule dedicated to parsing Pandas expressions.

    Args:
        expr_name (str): The name of the expression.
        next_keyword (str): The keyword following the expression or None
    """
    def __init__(self, expr_name, next_keyword):
        super().__init__(expr_name, next_keyword)
        self.operator_map["and"] = " & "
        self.operator_map["or"] = " | "
        self.operator_map["in"] = ".isin"
        self.operator_map["not"] = "~"

        self.operator_type["in"] = OperatorType.UNARY_FUNCTION


@grammar(docs.EXPRESSION_ONLY_DOC)
def expression_only(code, args):
    return code + args["expr"]


@grammar(docs.ON_DF_DOC)
def on_dataframe(code, args):
    return code + args["dataframe"]


@grammar(docs.CREATE_DF_DOC)
def create_dataframe(code, args, env):
    cols = list_to_string(args["header"])
    return code + env["import_name"] + ".DataFrame({}, columns={})".format(args["data"], cols)


@grammar(docs.LOAD_FROM_DOC)
def load_from(code, args, env):
    read_fun = ".read_csv(" if args["type"] == "csv" else ".read_json("
    return code + env["import_name"] + read_fun + args["path"] + ")"


@grammar(docs.SAVE_TO_DOC)
def save_to(code, args):
    read_fun = ".to_csv(" if args["type"] == "csv" else ".to_json("
    return code + read_fun + args["path"] + ")"


@grammar(docs.UNION_DOC)
def union(code, args, env):
    lhs, rhs = split_code(code, " = on ")
    return lhs + env["import_name"] + ".concat([{}, {}])".format(rhs, args["table"])


@grammar(docs.DIFFERENCE_DOC)
def difference(code, args):
    _, rhs = split_code(code, " = on ")
    return code + "[~{}.isin({}).all(1)]".format(rhs, args["table"])


@grammar(docs.INTERSECTION_DOC)
def intersection(code, args):
    return code + ".merge({})".format(args["table"])


@grammar(docs.SELECT_COLUMNS_DOC)
def select_columns(code, args):
    return code + "[{}]".format(list_to_string(args["columns"]))


@grammar(docs.SELECT_ROWS_DOC, PandasExpressionRule)
def select_rows(code, args):
    return code + "[{}]".format(args["condition"])


@grammar(docs.DROP_COLUMNS_DOC)
def drop_columns(code, args):
    return code + ".drop(columns={})".format(list_to_string(args["columns"]))


@grammar(docs.JOIN_DOC)
def join(code, args):
    cols = list_to_string(args["columns"])
    return code + ".join({}, on={}, how='{}')".format(args["table"], cols, args["how"])


@grammar(docs.GROUP_BY_DOC)
def group_by(code, args):
    cols = list_to_string(args["columns"])
    return code + ".groupby({}).{}()".format(cols, args["aggregation"])


@grammar(docs.REPLACE_VALUES_DOC)
def replace_values(code, args):
    return code + ".replace({}, {})".format(args["old_value"], args["new_value"])


@grammar(docs.APPEND_COLUMN_DOC)
def append_column(code, args):
    _, rhs = split_code(code, " = ")
    expr = re.sub(re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*\."), r"row.", args["col_expr"])
    col = rhs + ".apply(lambda row: {}, axis=1).values".format(expr)
    return code + ".assign(**{" + args["col_name"] + ": " + col + "})"


@grammar(docs.SORT_BY_DOC)
def sort_by(code, args):
    cols, order = zip(*args["columns"])
    cols = list_to_string(cols)
    order = [o == "ascending" for o in order]
    return code + ".sort_values({}, axis='index', ascending={})".format(cols, order)


@grammar(docs.DROP_DUPLICATES_DOC)
def drop_duplicates(code):
    return code + ".drop_duplicates()"


@grammar(docs.RENAME_COLUMNS_DOC)
def rename_columns(code, args):
    cols_map = "{" + ", ".join([k + ": " + v for k, v in args["columns"]]) + "}"
    return code + ".rename(columns={})".format(cols_map)


@grammar(docs.SHOW_DOC)
def show(code):
    return "print({})".format(code)


@grammar(docs.SHOW_SCHEMA_DOC)
def show_schema(code):
    return code + ".info(verbose=False)"


@grammar(docs.DESCRIBE_DOC)
def describe(code):
    return code + ".describe()"


@grammar(docs.HEAD_DOC)
def head(code, args):
    return code + ".head({})".format(args["num_rows"])


@grammar(docs.COUNT_DOC)
def count(code):
    return code + ".shape[0]"




PandasCodeGenerator.register_function(expression_only)
PandasCodeGenerator.register_function(on_dataframe)
PandasCodeGenerator.register_function(create_dataframe)
PandasCodeGenerator.register_function(load_from)
PandasCodeGenerator.register_function(save_to)
PandasCodeGenerator.register_function(union)
PandasCodeGenerator.register_function(difference)
PandasCodeGenerator.register_function(intersection)
PandasCodeGenerator.register_function(select_columns)
PandasCodeGenerator.register_function(select_rows)
PandasCodeGenerator.register_function(drop_columns)
PandasCodeGenerator.register_function(join)
PandasCodeGenerator.register_function(group_by)
PandasCodeGenerator.register_function(replace_values)
PandasCodeGenerator.register_function(append_column)
PandasCodeGenerator.register_function(sort_by)
PandasCodeGenerator.register_function(drop_duplicates)
PandasCodeGenerator.register_function(rename_columns)
PandasCodeGenerator.register_function(show)
PandasCodeGenerator.register_function(show_schema)
PandasCodeGenerator.register_function(describe)
PandasCodeGenerator.register_function(head)
PandasCodeGenerator.register_function(count)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""This module contains the CodeMap, CodeGenerator and Recommendation classes.

The CodeMap is essentially a dispatcher dict, which contains all dynamically added
grammar rules. It stores them in his __dict__ attribute. It implements the interface
of a python dictonary.

The CodeGenerator generates executable code from DSL statements.
In generall one want to derive from this class to crate a CodeGenerator with a specific
target, such a Pandas, Spark, R, etc.

A Recommendation is returned by the CodeGenerator instead of executable code in case no code
could be generated due to an incomplete or incorrect DSL argument list.

Example:
    First import::

        from nldsl import CodeGenerator, grammar

    Create custom code generator::

        class MyCodeGenerator(CodeGenerator):
            def __init__(my_env_var_1, ..., my_env_var_n=default_val):
                # all arguments passed to super.__int__() will be added to the environment,
                # which in turn will be passed to any dynamic grammar rule if requested.
                super().__init__(my_env_var_1=my_env_var_1, ..., my_env_var_n=my_env_var_n)

    Create a grammar rule, specifing the `env` parameter request the environment dict::

        @grammar
        def my_grammar_rule(code, args, env):
            \"\"\"My grammar rule doc...
            Grammar:
                parse my expression !expr
            \"\"\"
            return code + args["expr"] + env["my_var_1"]

    Add to the code generators class itself::

        MyCodeGenerator.register_function(my_grammar_rule)

    Or only add it to specific instance instead::

        code_gen = MyCodeGenerator(my_env_var_1)
        code_gen[\"parse my expression\"] = my_grammar_rule

    One may also tell the code generator to infer the name::

        code_gen[\"__infer__\"] = my_grammar_rule
"""

import re
import itertools as it
from collections.abc import MutableMapping
from os.path import join, abspath, dirname
from textx import metamodel_from_str
from .utils import cast_down, convert_function
from .internal import PipeFunction
from .rules import KeywordRule, VariableRule
from .types import Value, Keyword, Variable, Expression, VarList, is_value, is_keyword
from .exceptions import DSLFunctionError




class Recommendation:
    """Provides recommendation on how to continue with a DSL statement.

    A Recommendation is returned by the CodeGenerator instead of executable code in case no code
    could be generated due to an incomplete or incorrect DSL argument list.

    Args:
        rule (grammar rule): A Keyword-/Variable-/Expression-/VarListRule.
    """

    def __init__(self, rule, doc=""):
        """Please see help(Recommendation)."""
        self.rule = rule
        self.doc = doc


    @property
    def documentation(self):
        """The docstring of the grammar rule which raised this error."""
        return self.doc


    @property
    def description(self):
        """A string describing the arguments of the grammar rule which raised this error."""
        return self.rule.description


    def recommend(self):
        """Provide a list of recommendations on how to continue with a DSL statement.

        Returns:
            (list) Contains either a keyword, all choices for a variable or nothing.
        """
        if isinstance(self.rule, KeywordRule):
            return [self.rule.keyword]
        if isinstance(self.rule, VariableRule):
            return self.rule.choices
        return []




class CodeMap(MutableMapping):
    """A CodeMap maps the names of DSL functions to python functions.

    The names of DSL function are expected to be a sequence of lower case letters seperated
    by whitespaces. A CodeMap implements the dictonary interface.

    A new function may either be added to a specific instance of CodeMap::

        my_CodeMap["my dsl function name"] = my_function

    Or to the class itself::

        CodeMap.register_function(my_function, "my dsl function name")
    """

    CODEMAP_VERIFY_TAG = True # The CodeMap and all its child classes must have this tag!

    static_longest_name = 0
    infer_name_constant = "__infer__"
    dynamic_prefix = "_CODE_MAP_DYNAMIC_FUNCTION"
    dynamic_length = 4


    def __init__(self):
        """Please see help(CodeMap)."""
        self.own_longest_name = self._compute_own_longest_name()


    @property
    def longest_name(self):
        """The length of the longest sequence of tokens, which correspond to a function name."""
        return max(self.static_longest_name, self.own_longest_name)


    def __getitem__(self, name):
        """Access the function that is associated with `name` and attached to this instance.

        Args:
            name (str): A name corresponding to the desired function.

        Returns:
            (function) A python function, which takes a list as argument and returns a string.

        Raises:
            KeyError: If no function corresponding to `name` exists.
        """
        fun_name, _ = self._convert_name(name)
        if not hasattr(self, fun_name):
            raise KeyError("A function corresponding to the name `" + name + "` does not exists")
        return getattr(self, fun_name)


    def __setitem__(self, name, fun):
        """Define a function that will be associated with `name` only for this instance.

        Args:
            name (str): The name to be assoicated with the function.
            fun (function): The function this name should map to.

        Returns:
            (class CodeMap) this instance of the class
        """
        if name == self.infer_name_constant:
            name = fun.grammar_rule_name

        fun_name, length = self._convert_name(name)

        if length > self.own_longest_name:
            self.own_longest_name = length
        # calling setattr(<x>, name, <function>) will automatically
        # add the argument `self` to the prototype of <function> if <x> is an instance
        setattr(self, fun_name, convert_function(fun))
        return self


    def __delitem__(self, name):
        """Delete the function that is associated with `name` and attached to this instance.

        Args:
            name (str): A name corresponding to the desired function.

        Returns:
            (class CodeMap) this instance of the class.

        Raises:
            KeyError: If no function corresponding to `name` exists.
        """
        fun_name, length = self._convert_name(name)
        if not hasattr(self, fun_name):
            raise KeyError("A function corresponding to the name `" + name + "` does not exists")
        delattr(self, fun_name)
        if self.own_longest_name == length:
            self.own_longest_name = self._compute_own_longest_name()
        return self


    def __iter__(self):
        """Create an iterator over all dynmically added methods.

        Returns:
            (iterator) Said iterator.
        """
        # NOTE: When calling getattr(b, "__iter__")(b) b is a class not an instance!
        # This yields the correct results however because type(b) == abc.ABCMeta, hence ops == []
        # and b.__dict__ equals type(self).__dict if self would be an instance.
        # Meaning self_ops contains the result that would normal belong into ops and ops is empty
        bases = set(it.chain(*[getattr(b, "__iter__")(b) for b in self._determine_inheritance()]))
        offset = len(self.dynamic_prefix) + 1
        ops = [" ".join(k[offset:].split("_")) for k in type(self).__dict__ if self._is_dynamic(k)]
        self_ops = [" ".join(k[offset:].split("_")) for k in self.__dict__ if self._is_dynamic(k)]
        return it.chain(iter(self_ops + ops), bases)


    def __len__(self):
        """The number of all dynamically added methods.

        Returns:
            (int) The number of of dynamically added methods.
        """
        # See __iter__ impl for an explanation why this works.
        bases = set(it.chain(*[getattr(b, "__iter__")(b) for b in self._determine_inheritance()]))
        cls_len = len([k for k in type(self).__dict__ if self._is_dynamic(k)])
        self_len = len([k for k in self.__dict__ if self._is_dynamic(k)])
        return self_len + cls_len + len(bases)


    def update(self, other):
        """Updated the CodeMap with `other`.

        Args:
            other (CodeMap): Another instance of CodeMap.

        Returns:
            (CodeMap) This instance.
        """
        for name, fun in other.items():
            fun_name, _ = self._convert_name(name)
            # Do not overwrite dynamic class methods
            if not hasattr(type(self), fun_name):
                self[name] = fun
        return self


    def __eq__(self, other):
        """Compares two CodeMaps for equality.

        Args:
            other (CodeMap): Another instance of CodeMap

        Returns:
            (bool) True if the are equal, false otherwise.
        """
        if type(self) != type(other):
            return False
        if self.__dict__ != other.__dict__:
            return False
        return True


    @classmethod
    def request_function(cls, name):
        """Access the function that is associated with `name` for all instances of this class.

        Args:
            name (str): A name corresponding to the desired function.

        Returns:
            (function) A python function, or None.
        """
        fun_name, _ = cls._convert_name(name)
        if not hasattr(cls, fun_name):
            return None

        fun = getattr(cls, fun_name)

        def __anonymous_function_wrapper(code, args, env):
            return fun(cls, code, args, env)

        desc = fun.grammar_rule_desc if hasattr(fun, "grammar_rule_desc") else None
        gtype = fun.grammar_rule_type if hasattr(fun, "grammar_rule_type") else None
        __anonymous_function_wrapper.grammar_rule_desc = desc
        __anonymous_function_wrapper.grammar_rule_type = gtype
        __anonymous_function_wrapper.__name__ = fun.__name__
        __anonymous_function_wrapper.__doc__ = fun.__doc__

        return __anonymous_function_wrapper


    @classmethod
    def register_function(cls, fun, name=None):
        """Define a function that will be associated with `name` for all instances of this class.

        Args:
            fun (function): The function this name should map to.
            name (str): The name to be assoicated with the function.
        """
        if name is None:
            name = fun.grammar_rule_name

        fun_name, length = cls._convert_name(name)
        if length > cls.static_longest_name:
            cls.static_longest_name = length

        # calling setattr(<x>, name, <somefunction>) will NOT(!)
        # add the argument `self` to the prototype of <function> if <x> is a class
        tmp = convert_function(fun)

        def __anonymous_function_wrapper(self, code, args, env):
            return tmp(code, args, env)

        desc = fun.grammar_rule_desc if hasattr(fun, "grammar_rule_desc") else None
        gtype = fun.grammar_rule_type if hasattr(fun, "grammar_rule_type") else None
        __anonymous_function_wrapper.grammar_rule_desc = desc
        __anonymous_function_wrapper.grammar_rule_type = gtype
        __anonymous_function_wrapper.__name__ = fun.__name__
        __anonymous_function_wrapper.__doc__ = fun.__doc__

        setattr(cls, fun_name, __anonymous_function_wrapper)
        return cls


    @classmethod
    def remove_function(cls, name):
        """Delete the function that is associated with `name` from all instances of this class

        Args:
            name (str): A name corresponding to the desired function.

        Raises:
            KeyError: If no function corresponding to `name` exists.
        """
        fun_name, length = cls._convert_name(name)
        if not hasattr(cls, fun_name):
            raise KeyError("A function corresponding to the name `" + name + "` does not exists")
        delattr(cls, fun_name)
        if cls.static_longest_name == length:
            cls.static_longest_name = cls._compute_static_longest_name()
        return cls


    @classmethod
    def _is_dynamic(cls, name):
        """Check if method has been added dynamically.

        Args:
            name (str): The name of the method.

        Returns:
            (bool) True if the method has been added dynamically, false otherwise.
        """
        return name.startswith(cls.dynamic_prefix)


    @classmethod
    def _convert_name(cls, name):
        """Convert the `name` to a proper function name.

        Args:
            name (str): The name to be converted

        Returns:
            (tuple) The converted name and the number of tokens
        """
        if name == "":
            return cls.dynamic_prefix + "_", 0

        tokens = name.split()
        prefix = "" if tokens[0].startswith("_") else "_"
        return cls.dynamic_prefix + prefix + "_".join(tokens), len(tokens)


    def _compute_longest_name(self):
        """Compute the length of the longest sequence of tokens.

        Returns:
            (int): The computed length
        """
        return self._compute_own_longest_name() + self._compute_static_longest_name()


    def _compute_own_longest_name(self):
        """Compute the length of the longest sequence of tokens from this instance.

        Returns:
            (int): The computed length
        """
        # remove all methods, which are not dynamic grammar rules
        ops = [k[1:] for k in self.__dict__ if self._is_dynamic(k)]
        length = [len(s.split("_")) - self.dynamic_length for s in ops]
        return max(length) if length else 0


    @classmethod
    def _compute_static_longest_name(cls):
        """Compute the length of the longest sequence of tokens from this class.

        Returns:
            (int): The computed length
        """
        bases_sln = [getattr(b, "_compute_static_longest_name")()
                     for b in cls._determine_inheritance()]

        # remove all methods, which are not dynamic grammar rules
        ops = [k[1:] for k in cls.__dict__ if cls._is_dynamic(k)]
        length = [len(s.split("_")) - cls.dynamic_length for s in ops if s != ""] + bases_sln
        return max(length) if length else 0


    @classmethod
    def _determine_inheritance(cls):
        """Traverse the inheritance hierachy upwards and extract a set of all CodeMap-parents.

        Note:
            If one wants to access the member functions of the returned classes,
            it is necessary to use getitem(cls, "name of member function").

        Returns:
            (set): A set of classes derived from CodeMap and the CodeMap class itself.
        """
        bases = list(base for base in cls.__bases__ if hasattr(base, "CODEMAP_VERIFY_TAG"))
        bases += list(it.chain(*[base._determine_inheritance() for base in bases]))
        return set(bases)




class CodeGenerator(CodeMap):
    """A CodeGenerator translates DSL statements into executable code or changes the Grammer.

    There are two kind of DSL statements, the ones which can be evaluate to executable code
    and the ones, which extend the DSL Grammer.
    As a result parsing a set of DSL statements usually has two impacts.
    Executable code is generate and the CodeGenerator modifies itself in such a way
    that he is capable of parsing statements according to the new Grammer rules.

    Furthermore the CodeGenerator derives from the CodeMap class and his grammer can also
    be extended by rules, which can not be expressed within the DSL.
    This is done with the __setitem__ and register_function methods.

    Args:
        recommend (bool): Whether to return a Recommendation if possible or always raise an error.
        kwargs (dict): Specify the environment, which will be passed to all grammar rules.
    """

    # See nldsl/core/grammar/grammar.tx for a description of the available parameters.
    MODEL_TEMPLATE_PATH = join(abspath(dirname(__file__)), "grammar", "grammar.tx")

    EVAL_PIPE_PREFIX = "##"
    DEFINE_PIPE_PREFIX = "#$"

    # Call CodeGenerator.init_meta_model() to create the model regex & meta model
    MODEL_REGEX = None
    META_MODEL = None


    def __init__(self, recommend=True, **kwargs):
        """Please see help(CodeGenerator)."""
        super().__init__()
        self.recommend = recommend
        self.env = kwargs if kwargs is not None else {}


    def __call__(self, model, is_file=False):
        """Convert the given model into code.

        Args:
            model (str): A filename or string containg DSL statements
            is_file (bool): True if model is filename, False otherwise.

        Returns:
            (list) A list of strings containing executable code or recommendations.
        """
        results = sum([self._parse_input(m) for m in self._parse_model(model, is_file)], [])
        return self.postprocessing(results)



    @classmethod
    def register_dsl(cls, model, is_file=False):
        """Add all define statements in model to this class.

        Args:
            model (str): A filename or string containg DSL statements
            is_file (bool): True if model is filename, False otherwise.

        Returns:
            (CodeGenerator) This class.
        """
        stmts = sum([m.stmts for m in cls._parse_model(model, is_file)], [])
        for stmt in stmts:
            if stmt.define:
                cls._static_parse_define_pipe_stmt(stmt.define)
        return cls


    @classmethod
    def _parse_model(cls, models, is_file=False):
        """Convert the given model descriptions into a list of models

        Args:
            models (str): A filename or string containg DSL statements
            is_file (bool): True if model is filename, False otherwise.

        Returns:
            (list) A list of textx models, which contain only one stmt.
        """
        if is_file:
            with open(models, "r") as models_file:
                models = models_file.read()

        return [cls.META_MODEL.model_from_str(m) for m in cls.MODEL_REGEX.findall(models)]


    @classmethod
    def _static_parse_define_pipe_stmt(cls, node):
        """Parse a node corresponding to a definition statement.

        Instead of creating code, this statement will dynamically add a new grammer rule.
        This is done by construing a PipeFunction and adding it as an attribute to this class.

        Args:
            node (class DefinePipeStmt): The node describing the statement.

        Returns:
            (CodeGenerator) This instance.
        """
        name, instructions = cls._process_assignment(node.assign)

        # compute mapping from new pipeline to DSL statements
        fun_list = [cls._static_parse_def_op_expr(op) for op in node.operations]
        pipe_fun = PipeFunction(name, instructions, fun_list)
        cls.register_function(fun=pipe_fun, name=name)
        return cls


    @classmethod
    def _static_parse_def_op_expr(cls, node):
        """Parse a node corresponding to a definition operation expression.

        Args:
            node (class DefOpExpr): The node describing the expression

        Returns:
            (tuple) A tuple of a function and a list of its arguments

        Raises:
            ValueError: If the name of the function could not be resolved.
        """
        args = [cls._parse_def_arg_expr(arg) for arg in node.args]

        for num_tokens in range(cls.static_longest_name, -1, -1):
            tokens = [arg.data for arg in args[:num_tokens] if is_keyword(arg) or is_value(arg)]
            fun_name, name_length = " ".join(tokens), len(tokens)
            fun = cls.request_function(fun_name)
            if fun is not None:
                return (fun, args[name_length:])

        raise ValueError("Unable to resolve name of operation")


    def _parse_input(self, node):
        """Parse the AST rooted at `node`.

        Args:
            node (class Input): The root of the AST

        Returns:
            (list) A list of strings containing executable code or recommendations
        """
        for stmt in node.stmts:
            if stmt.define:
                self._parse_define_pipe_stmt(stmt.define)

        return [self._parse_eval_pipe_stmt(stmt.eval) for stmt in node.stmts if stmt.eval]


    def _parse_eval_pipe_stmt(self, node):
        """Parse a node corresponding to an evaluation statement.

        Args:
            node (class EvalPipeStmt): The node describing the statement.

        Returns:
            (str) A string containing executable code or a recommendation.
        """
        assign, code = node.result + " = " if node.result else "", ""
        for operation in node.operations:
            try:
                code = self._parse_op_expr(operation, code)
            except DSLFunctionError as err:
                if self.recommend:
                    return Recommendation(err.rule, err.doc)
                raise err from None
        return assign + code


    def _parse_define_pipe_stmt(self, node):
        """Parse a node corresponding to a definition statement.

        Instead of creating code, this statement will dynamically add a new grammer rule.
        This is done by construing a pipe_operator and adding as attribute to this instance
        of the class.

        Args:
            node (class DefinePipeStmt): The node describing the statement.

        Returns:
            (CodeGenerator) This instance.
        """
        name, instructions = self._process_assignment(node.assign)

        # compute mapping from new pipeline to DSL statements
        fun_list = [self._parse_def_op_expr(op) for op in node.operations]
        pipe_fun = PipeFunction(name, instructions, fun_list)
        self[name] = pipe_fun
        return self


    def _parse_op_expr(self, node, code):
        """Parse a node corresponding to an operation expression.

        Args:
            node (class OpExpr): The node describing the expression

        Returns:
            (str) A string containing a fragment of code.
        """
        args = [self._parse_arg_expr(arg) for arg in node.args]

        for num_tokens in range(self.longest_name, -1, -1):
            fun_name = " ".join([arg for arg in args[:num_tokens]])
            if fun_name in self:
                return self[fun_name](code, args[num_tokens:], self.env)

        raise ValueError("Unable to resolve name of operation")


    @classmethod
    def _parse_arg_expr(cls, node):
        """Parse a node corresponding to an argument expression.

        Args:
            node (class ArgExpr): The node describing the expression

        Returns:
            (str) A string containing a fragment of code.
        """
        return cls._parse_expr(node.expr)


    def _parse_def_op_expr(self, node):
        """Parse a node corresponding to a definition operation expression.

        Args:
            node (class DefOpExpr): The node describing the expression

        Returns:
            (tuple) A tuple of a function and a list of its arguments
        """
        args = [self._parse_def_arg_expr(arg) for arg in node.args]

        for num_tokens in range(self.longest_name, 0, -1):
            tokens = [arg.data for arg in args[:num_tokens] if is_keyword(arg) or is_value(arg)]
            fun_name, name_length = " ".join(tokens), len(tokens)
            if fun_name in self:
                return (self[fun_name], args[name_length:])

        raise ValueError("Unable to resolve name of operation")


    @classmethod
    def _parse_def_assign_expr(cls, node):
        """Parse a node corresponding to a definition assignment expression.

        Args:
            node (class DefAssignExpr): The node describing the expression

        Returns:
            (dict) A dict with a `data` and `type` field
        """
        if node.variable:
            return Variable(node.variable)
        if node.value:
            return Keyword(node.value)
        if node.def_expr:
            return Expression(node.def_expr)
        if node.var_list:
            var_list = [cls._parse_def_arg_list_expr(expr) for expr in node.var_list]
            return VarList(None, var_list)

        raise ValueError("Invalid assignment expression")


    @classmethod
    def _parse_def_arg_expr(cls, node):
        """Parse a node corresponding to a definition argument expression.

        Args:
            node (class DefArgExpr): The node describing the expression

        Returns:
            (dict) A dict with a `data` and `type` field
        """
        if node.variable:
            return Variable(node.variable)
        if node.expr:
            return Value(cls._parse_expr(node.expr))
        if node.def_expr:
            return Expression(node.def_expr)
        if node.var_list:
            var_list = [cls._parse_def_arg_list_expr(expr) for expr in node.var_list]
            return VarList(None, var_list)

        raise ValueError("Invalid argument expression")


    @classmethod
    def _parse_def_arg_list_expr(cls, node):
        """Parse a node corresponding to a definition argument list expression.

        Args:
            node (class DefArgExpr): The node describing the expression

        Returns:
            (dict) A dict with a `data` and `type` field
        """
        if node.variable:
            return Variable(node.variable)
        if node.value:
            return Keyword(node.value)
        if node.def_expr:
            raise ValueError("Argument list must not contain expression arguments.")
        raise ValueError("Invalid argument expression")


    @classmethod
    def _parse_expr(cls, node):
        """Parse a node corresponding to an expression.

        Args:
            node (class Expr): The node describing the expression

        Returns:
            (str) A string containing a fragment of code.
        """
        return cls._parse_value(node.value)


    @classmethod
    def _parse_value(cls, node):
        """Pare a node corresponding to a value.

        Args:
            node (TYPE): The node describing the value

        Returns:
            (str) A string containing the parsed value
        """
        if node.string:
            return "'" + node.string + "'"
        if node.float:
            return str(cast_down(node.float))
        return str(node.value)


    @classmethod
    def _process_assignment(cls, assignment):
        """Extract the name and instructions from an assignment definition

        Args:
            assignment (list): A list of DefAssignExpr objects

        Returns:
            (tuple) The name and instruction list.
        """
        # prepare assignment expressions
        assign = [cls._parse_def_assign_expr(ass) for ass in assignment]
        split_idx = [is_keyword(ass) for ass in assign].index(False)

        # process assignment expressions
        name = " ".join([ass.data for ass in assign[:split_idx]])
        instructions = assign[split_idx:]

        return name, instructions


    @classmethod
    def init_meta_model(
            cls,
            eval_prefix=EVAL_PIPE_PREFIX,
            define_prefix=DEFINE_PIPE_PREFIX
        ):
        """Compile the model template and create the meta model.

        Note:
            'eval_prefix' and 'define_prefix' must not be a valid
            sequence of tokens within the DSL nor a subset of such.

        Args:
            eval_prefix (str): The prefix for evaluation statements.
            define_prefix (str): The prefix for definition statements.

        Raises:
            ValueError: If any of the arguments are equal to each other.
        """
        if eval_prefix == define_prefix:
            raise ValueError("The eval and define prefix must not be equal.")

        with open(cls.MODEL_TEMPLATE_PATH, "r") as model_template_file:
            model_template = model_template_file.read()
            # See nldsl/core/grammar/grammar.tx for a description of the available parameters.
            model_template = model_template.replace("$EVAL_PIPE_PREFIX$", eval_prefix)
            model_template = model_template.replace("$DEFINE_PIPE_PREFIX$", define_prefix)

            cls.META_MODEL = metamodel_from_str(model_template)

        cls.EVAL_PIPE_PREFIX = eval_prefix
        cls.DEFINE_PIPE_PREFIX = define_prefix

        cls.MODEL_REGEX = re.compile(r"{eval}[^{define}]*|{define}[^{eval}]*".format(
            eval=re.escape(cls.EVAL_PIPE_PREFIX), define=re.escape(cls.DEFINE_PIPE_PREFIX)
        ))


    def postprocessing(self, code_list):
        """A hook for subclasses to modifiy the generate code list before it is returned.

        Args:
            code_list (list): A list of string containing executable code.

        Returns:
            (list) The modified code list
        """
        return code_list


    def extract_environment(self, source_lines):
        """A hook for sublcasses, which is used by the compiler and completion generator.

        Extract environment variables from the `source_lines` and update self.env accordingly.

        Args:
            source_lines (list): A list of strings containing source code.
        """


# Create the model regex & meta model
CodeGenerator.init_meta_model()



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

from setuptools import setup




LONG_DESC = "NLDSL is a tool to create domain specific languages (DSLs) for data science, \
which can be translated into executable code. A new DSL is created by deriving from \
the CodeGenerator class and rules are added to it via simple python functions. \
Besides providing code generation NLDSL allows the user to define DSL-level function, \
which are then treated as first-class rules. \
Currently we provide extensions for Pandas and PySpark."


setup(
    name="nldsl",
    version="0.15.3",
    url="https://gitlab.com/Einhornstyle/nldsl",
    project_urls={
        "Documentation": "https://einhornstyle.gitlab.io/nldsl/"
    },
    description="A DSL for data science with a syntax similar to a natural language.",
    long_description=LONG_DESC,
    author="Kevin Kiefer",
    author_email="abc.kiefer@gmail.com",
    install_requires=["textX"],
    packages=["nldsl", "nldsl.core"],
    package_data={"nldsl.core": ["grammar/*.tx"]},
    scripts=["scripts/nldsl-compile.py"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Free for non-commercial use",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Compilers"
    ],
    python_requires=">=3.4"
)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""For each DSL argument type exist a dedicated rule responsible for parsing it.

The class GrammarRule contains a list of such rules, which it uses to parse
a list of different DSL arguments. Amongst all rules the ExpressionRule is the
most sophistaced one and it is also intended that the a user may derive from it
to adjust the parsing of expressions to his/her needs.
"""

from collections import OrderedDict
from os.path import join, abspath, dirname
from textx import metamodel_from_file
from .utils import cast_down
from .types import is_keyword, is_variable, is_expression, is_var_list, OperatorType
from .exceptions import DSLArgumentError




class KeywordRule:
    """A KeywordRule is responsible for parsing keywords.

     Args:
        keyword (str): A string describing a keyword
    """

    def __init__(self, keyword):
        """Please see help(KeywordRule)."""
        self.keyword = keyword


    @property
    def name(self):
        """The name of the keyword."""
        return self.keyword


    @property
    def description(self):
        """A string containing a description of the grammar rule."""
        return self.name


    def __call__(self, res_dict, args):
        """Apply the KeywordRule.

        If the first element from `args` is a euqal to the keyword stored
        in this rule, it will be removed. Otherwise an error is raised.

        Args:
            res_dict (dict): redundant - exists only for api consistency
            args (list): A list of strings

        Returns:
            (tuple) The first element is `res_dict` and the second args

        Raises:
            ValueError: This error is raised if the keyword is not the next argument.
            DSLArgumentError: If the keyword is incomplete e.g ("keyw" instead of "keyword").
        """
        if args[0] != self.keyword:
            if self.keyword.startswith(args[0]):
                raise DSLArgumentError("Incomplete Keyword", self)
            err_str = "Invalid keyword - expected {} but got {} instead"
            raise ValueError(err_str.format(self.keyword, args))
        return res_dict, args[1:]




class VariableRule:
    """A VariableRule is responsible for parsing variables.

    Args:
        variable_name (str): A string containing the name of a variable
        choice_rules (dict): A dict mapping variable names to a list of possible values
    """

    def __init__(self, variable_name, choice_rules=None):
        """Please see help(VariableRule)."""
        choice_rules = {} if choice_rules is None else choice_rules
        self.variable_name = variable_name
        self.choices = choice_rules[variable_name] if variable_name in choice_rules else []


    @property
    def name(self):
        """The name of the variable."""
        return self.variable_name


    @property
    def description(self):
        """A string containing a description of the grammar rule."""
        return "$" + self.name


    def __call__(self, res_dict, args):
        """Apply the VariableRule.

        Adds the first value of `args` to the dictionary where the internally stored
        `variable_name` is used as a key. Raises an error if a list of choices for the variable
        is given (via `choice_rules`) and the value is not contained in that list.

        Args:
            res_dict (dict): A dictionary mapping variable names to values
            args (list): A list of strings

        Returns:
            (tuple) The first element is `res_dict` and the second args

        Raises:
            ValueError: Raised if the value is not contained in the list of choices.
        """
        if self.choices and not args[0] in self.choices:
            remaining_choices = [c for c in self.choices if c.startswith(args[0])]
            if remaining_choices:
                rule = VariableRule(self.name, {self.name: remaining_choices})
                raise DSLArgumentError("Incomplete choice", rule)
            raise ValueError("Invalid value for variable")
        res_dict[self.variable_name] = args[0]
        return res_dict, args[1:]




class VarListRule:
    """A VariableRule is responsible for parsing variable lists.

    Args:
        var_list_name (str): A string containing the name of a variable
        internal (list): List key_word_rules or VariableRules
        next_keyword (str): The keyword that comes after this varlist or None
    """

    def __init__(self, var_list_name, internal_rules, next_keyword=None):
        """Please see help(VarListRule)."""
        if var_list_name:
            self.var_list_name = var_list_name
            self._artificial_name = False
        else:
            name_list = [x.variable_name for x in internal_rules if isinstance(x, VariableRule)]
            self.var_list_name = tuple(name_list) if len(name_list) > 1 else name_list[0]
            self._artificial_name = True

        self.next_keyword = next_keyword
        self.internal_rules = internal_rules


    @property
    def name(self):
        """The name of the variable list."""
        return self.var_list_name


    @property
    def description(self):
        """A string containing a description of the grammar rule."""
        internal_desc = [r.description for r in self.internal_rules]
        name = "" if self._artificial_name else self.name
        return "$" + name + "[{}]".format(" ".join(internal_desc))


    def __call__(self, res_dict, args):
        """Apply the VariableRule.

        Adds a list to the dictionary with tuples containing
        the values of the variables in the list.

        Args:
            res_dict (dict): A dictionary mapping variable names to values
            args (list): A list of strings

        Returns:
            (tuple) The first element is `res_dict` and the second the remaining args

        Raises:
            DSLArgumentError: If a DSL argument is missing.
        """
        try:
            stop_idx = len(args) if self.next_keyword is None or self.next_keyword not in args\
                                 else args.index(self.next_keyword)
        except ValueError as err:
            next_rule = KeywordRule(self.next_keyword)
            raise DSLArgumentError("DSL keyword delimiter is missing", next_rule) from err

        res_dict[self.var_list_name] = []

        for _ in range(0, stop_idx, len(self.internal_rules)):
            values = []
            for rule in self.internal_rules:
                if not args:
                    raise DSLArgumentError("DSL argument is missing", rule)

                tmp_res, args = rule({}, args)
                if rule.name in tmp_res:
                    values.append(tmp_res[rule.name])

            values = tuple(values) if len(values) > 1 else values[0]
            res_dict[self.var_list_name].append(values)

        return res_dict, args


    def _apply_internal_rule(self, rule_idx, arg):
        """Apply the `rule_idx`'th internal rule to arg

        Args:
            rule_idx (int): The index of the rule
            arg (str): The argument to the rule

        Returns:
            (str): Either arg itself or None (depending on the rule)
        """
        rule = self.internal_rules[rule_idx]
        if isinstance(rule, KeywordRule):
            return None
        res, _ = rule({}, [arg])
        return res[rule.variable_name]




class ExpressionRule:
    """An ExpressionRule is responsible for parsing expression.

    Args:
        expr_name (str): The name of the expression.
        next_keyword (str): The keyword after this expression or None
    """

    meta_model = metamodel_from_file(join(abspath(dirname(__file__)),
                                          "grammar", "expr_grammar.tx"))

    def __init__(self, expr_name, next_keyword=None):
        """Please see help(ExpressionRule)."""
        self.expr_name = expr_name
        self.next_keyword = next_keyword
        self.operator_map = {
            "or": " or ", "and": " and ", "not": "not ",
            "+": " + ", "-": " - ", "*": " * ", "/": " / ", "%": " % ", "**": "**",
            "<": " < ", ">": " > ", "==": " == ", "!=": " != ",
            "<=": " <= ", ">=": " >= ", "<>": " <> ",
            "in": " in ", "not in": " not in ", "is": " is ", "is not": " is not ",
            "[": "[", "]": "]", "(": "(", ")": ")"
        }
        self.operator_type = {op: OperatorType.OPERATOR for op in self.operator_map}
        self.brackets = "()"


    @property
    def name(self):
        """The name of the expression."""
        return self.expr_name


    @property
    def description(self):
        """A string containing a description of the grammar rule."""
        return "!" + self.name


    def __call__(self, res_dict, args):
        """Apply the ExpressionRule.

        Adds a string containing the parsed expression to 'res_dict'

        Args:
            res_dict (dict): A dictionary mapping variable names to values
            args (list): A list of strings

        Returns:
            (tuple) The first element is `res_dict` and the second the remaining args
        """
        try:
            stop_idx = len(args) if self.next_keyword is None else args.index(self.next_keyword)
        except ValueError as err:
            next_rule = KeywordRule(self.next_keyword)
            raise DSLArgumentError("DSL keyword delimiter is missing", next_rule) from err

        try:
            model = self.meta_model.model_from_str(" ".join(args[:stop_idx]))
        except:
            raise DSLArgumentError("Incomplete expression", self)
        res_dict[self.expr_name] = self._parse_or_expr(model)
        return res_dict, args[stop_idx:]


    def _parse_or_expr(self, node):
        """Pare a node corresponding to an boolean or-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        exprs = [("or", self._parse_and_expr(expr)) for expr in node.exprs]
        return self.apply_operator_seq(exprs[0][1], exprs[1:])


    def _parse_and_expr(self, node):
        """Pare a node corresponding to an boolean and-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        exprs = [("and", self._parse_not_expr(expr)) for expr in node.exprs]
        return self.apply_operator_seq(exprs[0][1], exprs[1:])


    def _parse_not_expr(self, node):
        """Pare a node corresponding to an boolean not-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        if node.comp:
            return self._parse_comparison_expr(node.comp)
        exprs = [("not", self._parse_not_expr(expr)) for expr in node.exprs]
        return self.apply_operator_seq("", exprs)


    def _parse_comparison_expr(self, node):
        """Pare a node corresponding to a comparsion expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        rhs_list = [self._parse_comparison_expr_rhs(expr) for expr in node.rhs]
        return self.apply_operator_seq(self._parse_arith_expr(node.lhs), rhs_list)


    def _parse_comparison_expr_rhs(self, node):
        """Pare a node corresponding to the right hand side of comparisoin-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (tuple) The operator and the expression.
        """
        if node.operator == "notin":
            node.operator = "not in"
        if node.operator == "isnot":
            node.operator = "is not"
        return (node.operator, self._parse_arith_expr(node.expr))


    def _parse_arith_expr(self, node):
        """Pare a node corresponding to an arithmetic-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        rhs_list = [self._parse_arith_expr_rhs(expr) for expr in node.rhs]
        return self.apply_operator_seq(self._parse_term_expr(node.lhs), rhs_list)


    def _parse_arith_expr_rhs(self, node):
        """Pare a node corresponding to the right hand side of a arithmetic-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (tuple) The operator and the expression.
        """
        return (node.operator, self._parse_term_expr(node.expr))


    def _parse_term_expr(self, node):
        """Pare a node corresponding to a term-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        rhs_list = [self._parse_term_expr_rhs(expr) for expr in node.rhs]
        return self.apply_operator_seq(self._parse_factor_expr(node.lhs), rhs_list)


    def _parse_term_expr_rhs(self, node):
        """Pare a node corresponding to the right hand side of term-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (tuple) The operator and the expression.
        """
        return (node.operator, self._parse_factor_expr(node.expr))


    def _parse_factor_expr(self, node):
        """Pare a node corresponding to an arithmetic factor-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        if node.expr:
            return self._parse_power_expr(node.expr)

        operator = self.operator_map[node.operator]
        operator = self.factor_sign(operator)

        if self.operator_type[node.operator] == OperatorType.UNARY_FUNCTION:
            return self.apply_operator(operator, rhs=self._parse_factor_expr(node.factor))

        return operator + self._parse_factor_expr(node.factor)


    def _parse_power_expr(self, node):
        """Pare a node corresponding to an arithmetic power-expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        if node.factor:
            lhs, rhs = self._parse_atom_expr(node.atom), self._parse_factor_expr(node.factor)
            return self.apply_operator("**", lhs, rhs)
        return self._parse_atom_expr(node.atom)


    def _parse_atom_expr(self, node):
        """Pare a node corresponding to an atomic expression

        Args:
            node (OrExpr): The node describing the expression

        Returns:
            (str) A string containing the parsed expression
        """
        if node.expr:
            return self._parse_or_expr(node.expr)
        if node.dyck_expr:
            left, right = self.operator_map["("], self.operator_map[")"]
            return left + self._parse_or_expr(node.dyck_expr) + right
        if node.value_list:
            left, right = self.operator_map["["], self.operator_map["]"]
            vals = [self._parse_value(val) for val in node.value_list]
            return left + ", ".join(vals) + right
        return self._parse_value(node.value)


    @classmethod
    def _parse_value(cls, node):
        """Pare a node corresponding to a value.

        Args:
            node (TYPE): The node describing the value

        Returns:
            (str) A string containing the parsed value
        """
        if node.string:
            return "'" + node.string + "'"
        if node.float:
            return str(cast_down(node.float))
        return str(node.value)


    def apply_operator_seq(self, lhs, rhs_list):
        """Concatenates `lhs` and all expressions in `rhs_list`.

        Args:
            lhs (str): The left most expression
            rhs_list (list): A list of (operator, expression) tuples

        Returns:
            (str) A string containing the parsed expressions
        """
        result = lhs
        for operation, rhs in rhs_list:
            result = self.apply_operator(operation, result, rhs)
        return result


    def apply_operator(self, operator, lhs="", rhs=""):
        """Applies the `operator` to the left- and right-hand-side expression

        If the operator is a function the expressions will become its arguments.
        Otherwise they will simply be concatenated.

        Args:
            operator (str): A string containg an operator
            lhs (str): A string containg an expression
            rhs (str): A string containg an expression

        Returns:
            (str) The resulting expression
        """
        sym = self.operator_map[operator]
        if self.operator_type[operator] == OperatorType.UNARY_FUNCTION:
            return lhs + sym + self.brackets[0] + rhs + self.brackets[1]
        if self.operator_type[operator] == OperatorType.BINARY_FUNCTION:
            return sym + self.brackets[0] + lhs + ", " + rhs + self.brackets[1]
        return lhs + sym + rhs


    def factor_sign(self, operator):
        """Modify an operator, which describes the sign of a base.

        The default implementation simply strips all whitespaces.

        Args:
            operator (str): A string representation of the operator.

        Returns:
            (str) The operator string stripped from preceeding and trailing whitespaces.
        """
        return operator.strip()




class GrammarRule:
    """A GrammarRule consists out of a list of rules.

    Args:
        instructions (list): Descibes how to synthesize the rule.
        choices (dict): A dict, which maps variable names to values.
        expr_rule (ExpressionRule): The type of expression rule to be used.
    """

    def __init__(self, instructions, choices=None, expr_rule=ExpressionRule):
        """Please see help(GrammarRule)."""
        choices = {} if choices is None else choices
        self._expr_rule = expr_rule
        self._rules = self._parse_rules(instructions, choices)
        self._choices = self._parse_choices(self._rules, choices)


    @property
    def rules(self):
        """A list of the internal rules."""
        return self._rules

    @property
    def choices(self):
        """A dictionary wich maps variable names to lists of choices."""
        return self._choices

    @property
    def expression_rule(self):
        """The type of expression rule that is used."""
        return self._expr_rule

    @property
    def description(self):
        """A string containing a description of the grammar rule."""
        return " ".join(r.description for r in self.rules)


    def __call__(self, args):
        """Convert the list of arguments `args` into a dictinonary.

        All keywords are removed from `args`.
        All variable names become keys to their corresponding values.
        VarLists are converted into lists of tuples, where each tuples
        contains the values of the variables contained in the list in order.

        Args:
            args (list): A list of strings.

        Returns:
            (dict) A dictionary, which maps from variable names to values.

        Raises:
            ValueError: Raised if too many arguments are provided.
            DSLArgumentError: If a DSL argument is missing
        """
        res_dict = OrderedDict()
        for rule in self.rules:
            if not args:
                raise DSLArgumentError("DSL argument is missing", rule)
            res_dict, args = rule(res_dict, args)

        if args:
            raise ValueError("Too many arguments")
        return res_dict


    def _parse_rules(self, instructions, choices):
        """Parse the grammar rules as specified by the instructions.

        Args:
            instructions (list): Descibes how to synthesize the rule.
            choices (dict): A dict, which maps variable names to values.

        Returns:
            (list) A list of functors, which implement the rules.
        """
        rules = []
        for idx, arg in enumerate(instructions):
            next_arg = instructions[idx+1] if idx < len(instructions) - 1 else None
            # variable list parsing rule
            if is_var_list(arg):
                rules.append(self._create_var_list_rule(arg, next_arg, choices))
            # expression parsing rule
            elif is_expression(arg):
                rules.append(self._create_expression_rule(arg, next_arg))
            # variable parsing rule
            elif is_variable(arg):
                rules.append(VariableRule(arg.data, choices))
            # keyword parsing rule
            elif is_keyword(arg):
                rules.append(KeywordRule(arg.data))
            else:
                raise ValueError("Unkown rule type")

        return rules


    def _parse_choices(self, rules, choices):
        """Parse the lists of choices for all rules.

        Args:
            rule (list): A list of rules.
            choices (dict): The dictionary of already extracted choices.

        Returns:
            (dict): A mapping from variables names to lists of choices if any.
        """
        for rule in rules:
            if isinstance(rule, VariableRule) and rule.choices:
                choices[rule.name] = rule.choices
            if isinstance(rule, VarListRule):
                choices = self._parse_choices(rule.internal_rules, choices)
        return choices


    def _create_var_list_rule(self, arg, next_arg, choices):
        """ Create an instance of VarListRule.

        Args:
            arg (dict): The argument describing the variable list rule.
            next_arg (dict): The next argument in the instructions or None.
            choices (dict): A map from variable names to values.

        Returns:
            (VarListRule) A new variable list rule.

        Raises:
            ValueError: If the VarList is neither the last argument nor followed by a keyword.
        """
        # internals of a variable list are parsed recursively
        internal_rules = self._parse_rules(arg.data, choices)

        # ensure the next_arg does not exist or is a keyword
        if not next_arg is None and not is_keyword(next_arg):
            raise ValueError("A variable list must be followed by a keyword or be the last rule.")

        next_arg = next_arg.data if not next_arg is None else next_arg
        return VarListRule(arg.name, internal_rules, next_arg)


    def _create_expression_rule(self, arg, next_arg):
        """Create an instance of ExpressionRule.

        Args:
            arg (dict): The argument describing the expression rule.
            next_arg (dict): The next argument in the instructions or None.

        Returns:
            (ExpressionRule) A new variable list rule.

        Raises:
            ValueError: If the Expression is neither the last argument nor followed by a keyword.
        """
        # ensure the next_arg does not exist or is a keyword
        if not next_arg is None and not is_keyword(next_arg):
            raise ValueError("An expression must be followed by a keyword or be the last rule.")

        next_arg = next_arg.data if not next_arg is None else next_arg
        return self.expression_rule(arg.data, next_arg)



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

""" NLDSL is a tool to create DSLs for data science, which can be translated into executable code.

A new DSL is created by deriving from the CodeGenerator class and rules are added to it via simple
python functions.
Besides providing code generation NLDSL allows the user to define DSL-level function,
which are then treated as first-class rules.
Currently we provide extensions for Pandas and PySpark.

Example:
    We are going to use the PandasCodeGenerator for this example::

        from nldsl import PandasCodeGenerator, grammar

        code_gen = PandasCodeGenerator

    Define a DSL statement and create executable code::

        dsl_stmt = "## x = on df | rename columns old1 to new1, old2 to new2"
        python_code = code_gen(dsl_stmt)

    Add a new rule to the DSL and use it to create executable code::

        new_rule = "#$ my rule $x = rename columns old to $x | drop duplicates"
        dsl_stmt = "## x = on df | my rule new"
        python_code = code_gen(new_rule + dsl_stmt)

    Implement a new rule via python::

        def rename_columns(code, args):
            col_map = {args[i]:args[i+2] for i in range(0, len(args), 3)}
            return code + ".rename(columns={})".format(col_map)

    Implement the "rename columns" rule using the @grammar decorator::

        @grammar
        def rename_columns(code, args):
            \"\"\"Desc...
            Grammar:
	            rename columns $my_var[$old to $new]
            \"\"\"
            return code + ".rename(columns={})".format(dict(args["my_var"]))

    New rule can either be added to a certain instance of code_generator or to the class itself::

        code_gen["rename columns"] = rename_columns
        PandasCodeGenerator.register_function("rename colummns", rename_columns)
"""

from .core import CodeGenerator, Recommendation, grammar
from .pandas_extension import PandasCodeGenerator
from .spark_extension import SparkCodeGenerator



# NLDSL (c) 2019 by Kevin Kiefer <abc.kiefer@gmail.com>, Heidelberg University
#
# NLDSL is licensed under a
# Creative Commons Attribution-NonCommercial 3.0 Unported License.
#
# You should have received a copy of the license along with this
# work.  If not, see <http://creativecommons.org/licenses/by-nc/3.0/>.

"""The `utils` module provides utility functions for the `core` package."""

from inspect import signature
from functools import wraps




def decorator_factory(decorator):
    """A decorator which improves the parsing of optional arguments of other decorators.

    Let a decorator of the following form be given::

        decorator(arg1=value1, ... , kwargsN=valueN)'

    One would need to write the following in case all arguments should be left as default::

        @decorator()
        def myFunction ...

    If we declare our decorator as follow::

        @decorator_factory
        def decorator(fun, arg1=value1, ... , kwargsN=valueN)

    We could write instead::

        @decorator
        def myFunction ...

    Furthermore, we do not have to write the boilderplate code necessary to implement
    a decorator with arguments in the first place.
    """
    @wraps(decorator)
    def decorator_wrapper(*args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])
        return lambda fun: decorator(fun, *args, **kwargs)
    return decorator_wrapper




def cast_down(value):
    """Convert an object to an integer if no rounding is required.

    If the object is a float, which can be converted to an integer
    without rounding int(`value`) is returned. Otherwise `value` is returned.

    Args:
        value (object): An object to possibly converted

    Returns:
       (object or int) Either `value` or its integer representation.
    """
    return int(value) if isinstance(value, float) and int(value) == value else value




def convert_function(fun):
    """Convert the function signature if necessary

    The target signature is:
    my_function(str: code, list: args, dict: env) -> str

    The changes are strictly additive, in that no argument will be removed.

    Args:
        fun (function): The function to be converterd.

    Returns:
        (function) A function with proper signature.

    Raises:
        ValueError: If `fun` has four or more parameters.
    """
    sig = signature(fun)
    num_params = len(sig.parameters)

    if num_params == 3:
        if not hasattr(fun, "grammar_rule_desc"):
            fun.grammar_rule_desc = None
        if not hasattr(fun, "grammar_rule_type"):
            fun.grammar_rule_type = None
        return fun

    if num_params == 2 and "args" in sig.parameters:
        def __anonymous_function_wrapper(code, args, env):
            return fun(code, args)
    elif num_params == 2 and "env" in sig.parameters:
        def __anonymous_function_wrapper(code, args, env):
            return fun(code, env)
    elif num_params == 1:
        def __anonymous_function_wrapper(code, args, env):
            return fun(code)
    else:
        raise ValueError("Invalid function signatur - got {}, expected {}"
                         .format(str(sig), "(code, args, env)"))

    desc = fun.grammar_rule_desc if hasattr(fun, "grammar_rule_desc") else None
    gtype = fun.grammar_rule_type if hasattr(fun, "grammar_rule_type") else None
    __anonymous_function_wrapper.grammar_rule_desc = desc
    __anonymous_function_wrapper.grammar_rule_type = gtype
    __anonymous_function_wrapper.__name__ = fun.__name__
    __anonymous_function_wrapper.__doc__ = fun.__doc__
    return __anonymous_function_wrapper




def split_code(code, split_on=" = "):
    """Split the code in half on the string specified by `split_on`.

    Args:
        code (str): A string containg code
        split_on (str): The string on which to split the code

    Returns:
        (tuple) Code before and after the string `split_on`
    """
    splitted_code = code.split(split_on)

    if len(splitted_code) > 2:
        raise ValueError("Code has been splitted into more than two fragements")

    return splitted_code if len(splitted_code) == 2 else ("", splitted_code[0])




def list_to_string(string_list):
    """Converts a list of strings into a single string.

    Unlike {}.format(some_list) this treats elements of the form "'...'" properly.

    Args:
        string_list (list): A list of strings

    Returns:
        (str) A string containing the list.
    """
    return "[" + ", ".join(string_list) + "]"
